# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import sys
import os
import torch
import nimble
from typing import List, Dict, Any, Optional
from rich.prompt import Confirm, Prompt, PromptBase
import requests
from dataclasses import dataclass
from . import defaults

console = nimble.__console__


class IntListPrompt(PromptBase):
    """Prompt for a list of integers."""

    def check_choice(self, value: str) -> bool:
        assert self.choices is not None
        # check if value is a valid choice or all the values in a list of ints are valid choices
        return (
            value == "All"
            or value in self.choices
            or all(
                val.strip() in self.choices
                for val in value.replace(",", " ").split()
            )
        )


def check_netuid_set(
    config: "nimble.config",
    nbnetwork: "nimble.nbnetwork",
    allow_none: bool = False,
):
    if nbnetwork.network != "nakamoto":
        all_netuids = [str(netuid) for netuid in nbnetwork.get_cosmos()]
        if len(all_netuids) == 0:
            console.print(":cross_mark:[red]There are no open networks.[/red]")
            sys.exit()

        # Make sure netuid is set.
        if not config.is_set("netuid"):
            if not config.no_predict:
                netuid = IntListPrompt.ask(
                    "Enter netuid",
                    choices=all_netuids,
                    default=str(all_netuids[0]),
                )
            else:
                netuid = str(defaults.netuid) if not allow_none else "None"
        else:
            netuid = config.netuid

        if (
            isinstance(netuid, str)
            and netuid.lower() in ["none"]
            and allow_none
        ):
            config.netuid = None
        else:
            if isinstance(netuid, list):
                netuid = netuid[0]
            try:
                config.netuid = int(netuid)
            except:
                raise ValueError(
                    'netuid must be an integer or "None" (if applicable)'
                )


def check_for_cuda_reg_config(config: "nimble.config") -> None:
    """Checks, when CUDA is available, if the user would like to register with their CUDA device."""
    if torch.cuda.is_available():
        if not config.no_predict:
            if config.pow_register.cuda.get("use_cuda") == None:  # flag not set
                # Ask about cuda registration only if a CUDA device is available.
                cuda = Confirm.ask(
                    "Detected CUDA device, use CUDA for registration?\n"
                )
                config.pow_register.cuda.use_cuda = cuda

            # Only ask about which CUDA device if the user has more than one CUDA device.
            if (
                config.pow_register.cuda.use_cuda
                and config.pow_register.cuda.get("dev_id") is None
            ):
                devices: List[str] = [
                    str(x) for x in range(torch.cuda.device_count())
                ]
                device_names: List[str] = [
                    torch.cuda.get_device_name(x)
                    for x in range(torch.cuda.device_count())
                ]
                console.print("Available CUDA devices:")
                choices_str: str = ""
                for i, device in enumerate(devices):
                    choices_str += "  {}: {}\n".format(device, device_names[i])
                console.print(choices_str)
                dev_id = IntListPrompt.ask(
                    "Which GPU(s) would you like to use? Please list one, or comma-separated",
                    choices=devices,
                    default="All",
                )
                if dev_id.lower() == "all":
                    dev_id = list(range(torch.cuda.device_count()))
                else:
                    try:
                        # replace the commas with spaces then split over whitespace.,
                        # then strip the whitespace and convert to ints.
                        dev_id = [
                            int(dev_id.strip())
                            for dev_id in dev_id.replace(",", " ").split()
                        ]
                    except ValueError:
                        console.log(
                            ":cross_mark:[red]Invalid GPU device[/red] [bold white]{}[/bold white]\nAvailable CUDA devices:{}".format(
                                dev_id, choices_str
                            )
                        )
                        sys.exit(1)
                config.pow_register.cuda.dev_id = dev_id
        else:
            # flag was not set, use default value.
            if config.pow_register.cuda.get("use_cuda") is None:
                config.pow_register.cuda.use_cuda = (
                    defaults.pow_register.cuda.use_cuda
                )


def get_hotkey_wallets_for_wallet(wallet) -> List["nimble.wallet"]:
    hotkey_wallets = []
    hotkeys_path = wallet.path + "/" + wallet.name + "/hotkeys"
    try:
        hotkey_files = next(os.walk(os.path.expanduser(hotkeys_path)))[2]
    except StopIteration:
        hotkey_files = []
    for hotkey_file_name in hotkey_files:
        try:
            hotkey_for_name = nimble.wallet(
                path=wallet.path, name=wallet.name, hotkey=hotkey_file_name
            )
            if (
                hotkey_for_name.hotkey_file.exists_on_device()
                and not hotkey_for_name.hotkey_file.is_encrypted()
            ):
                hotkey_wallets.append(hotkey_for_name)
        except Exception:
            pass
    return hotkey_wallets


def get_coldkey_wallets_for_path(path: str) -> List["nimble.wallet"]:
    try:
        wallet_names = next(os.walk(os.path.expanduser(path)))[1]
        return [nimble.wallet(path=path, name=name) for name in wallet_names]
    except StopIteration:
        # No wallet files found.
        wallets = []
    return wallets


def get_all_wallets_for_path(path: str) -> List["nimble.wallet"]:
    all_wallets = []
    cold_wallets = get_coldkey_wallets_for_path(path)
    for cold_wallet in cold_wallets:
        if (
            cold_wallet.coldkeypub_file.exists_on_device()
            and not cold_wallet.coldkeypub_file.is_encrypted()
        ):
            all_wallets.extend(get_hotkey_wallets_for_wallet(cold_wallet))
    return all_wallets


@dataclass
class DelegatesDetails:
    name: str
    url: str
    description: str
    signature: str

    @classmethod
    def from_json(cls, json: Dict[str, any]) -> "DelegatesDetails":
        return cls(
            name=json["name"],
            url=json["url"],
            description=json["description"],
            signature=json["signature"],
        )


def _get_delegates_details_from_github(
    requests_get, url: str
) -> Dict[str, DelegatesDetails]:
    response = requests_get(url)

    if response.status_code == 200:
        all_delegates: Dict[str, Any] = response.json()
        all_delegates_details = {}
        for delegate_hotkey, delegates_details in all_delegates.items():
            all_delegates_details[delegate_hotkey] = DelegatesDetails.from_json(
                delegates_details
            )
        return all_delegates_details
    else:
        return {}


def get_delegates_details(url: str) -> Optional[Dict[str, DelegatesDetails]]:
    try:
        return _get_delegates_details_from_github(requests.get, url)
    except Exception:
        return None  # Fail silently



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import nimble

from starlette.responses import StreamingResponse as _StreamingResponse
from starlette.responses import Response
from starlette.types import Send, Receive, Scope
from typing import Callable, Awaitable, List
from pydantic import BaseModel
from abc import ABC, abstractmethod


class NBStreamingResponseModel(BaseModel):
    """
    NBStreamingResponseModel is a Pydantic model that encapsulates the token streamer callable for Pydantic validation.
    It is used within the StreamingNucleon class to create a NBStreamingResponse object, which is responsible for handling
    the streaming of tokens.

    The token streamer is a callable that takes a send function and returns an awaitable. It is responsible for generating
    the content of the streaming response, typically by processing tokens and sending them to the client.

    This model ensures that the token streamer conforms to the expected signature and provides a clear interface for
    passing the token streamer to the NBStreamingResponse class.

    Attributes:
        token_streamer: Callable[[Send], Awaitable[None]]
            The token streamer callable, which takes a send function (provided by the ASGI server) and returns an awaitable.
            It is responsible for generating the content of the streaming response.
    """

    token_streamer: Callable[[Send], Awaitable[None]]


class StreamingNucleon(nimble.Nucleon, ABC):
    """
    The StreamingNucleon class is designed to be subclassed for handling streaming responses in the nimble network.
    It provides abstract methods that must be implemented by the subclass to deserialize, process streaming responses,
    and extract JSON data. It also includes a method to create a streaming response object.
    """

    class Config:
        validate_assignment = True

    class NBStreamingResponse(_StreamingResponse):
        """
        NBStreamingResponse is a specialized subclass of the Starlette StreamingResponse designed to handle the streaming
        of tokens within the nimble network. It is used internally by the StreamingNucleon class to manage the response
        streaming process, including sending headers and calling the token streamer provided by the subclass.

        This class is not intended to be directly instantiated or modified by developers subclassing StreamingNucleon.
        Instead, it is used by the create_streaming_response method to create a response object based on the token streamer
        provided by the subclass.

        Reference: https://www.starlette.io/responses/#streamingresponse
        """

        def __init__(self, model: NBStreamingResponseModel, **kwargs):
            """
            Initializes the NBStreamingResponse with the given token streamer model.

            Args:
                model: A NBStreamingResponseModel instance containing the token streamer callable, which is responsible
                    for generating the content of the response.
                **kwargs: Additional keyword arguments passed to the parent StreamingResponse class.
            """
            super().__init__(content=iter(()), **kwargs)
            self.token_streamer = model.token_streamer

        async def stream_response(self, send: Send):
            """
            Asynchronously streams the response by sending headers and calling the token streamer.

            This method is responsible for initiating the response by sending the appropriate headers, including the
            content type for event-streaming. It then calls the token streamer to generate the content and sends the
            response body to the client.

            Args:
                send: A callable to send the response, provided by the ASGI server.
            """
            headers = [
                (b"content-type", b"text/event-stream")
            ] + self.raw_headers

            await send(
                {
                    "type": "http.response.start",
                    "status": 200,
                    "headers": headers,
                }
            )

            await self.token_streamer(send)

            await send(
                {"type": "http.response.body", "body": b"", "more_body": False}
            )

        async def __call__(self, scope: Scope, receive: Receive, send: Send):
            """
            Asynchronously calls the stream_response method, allowing the NBStreamingResponse object to be used as an ASGI
            application.

            This method is part of the ASGI interface and is called by the ASGI server to handle the request and send the
            response. It delegates to the stream_response method to perform the actual streaming process.

            Args:
                scope: The scope of the request, containing information about the client, server, and request itself.
                receive: A callable to receive the request, provided by the ASGI server.
                send: A callable to send the response, provided by the ASGI server.
            """
            await self.stream_response(send)

    @abstractmethod
    async def process_streaming_response(self, response: Response):
        """
        Abstract method that must be implemented by the subclass.
        This method should provide logic to handle the streaming response, such as parsing and accumulating data.
        It is called as the response is being streamed from the network, and should be implemented to handle the specific
        streaming data format and requirements of the subclass.

        Args:
            response: The response object to be processed, typically containing chunks of data.
        """
        ...

    @abstractmethod
    def extract_response_json(self, response: Response) -> dict:
        """
        Abstract method that must be implemented by the subclass.
        This method should provide logic to extract JSON data from the response, including headers and content.
        It is called after the response has been processed and is responsible for retrieving structured data
        that can be used by the application.

        Args:
            response: The response object from which to extract JSON data.
        """
        ...

    def create_streaming_response(
        self, token_streamer: Callable[[Send], Awaitable[None]]
    ) -> NBStreamingResponse:
        """
        Creates a streaming response using the provided token streamer.
        This method can be used by the subclass to create a response object that can be sent back to the client.
        The token streamer should be implemented to generate the content of the response according to the specific
        requirements of the subclass.

        Args:
            token_streamer: A callable that takes a send function and returns an awaitable. It's responsible for generating the content of the response.

        Returns:
            NBStreamingResponse: The streaming response object, ready to be sent to the client.
        """
        model_instance = NBStreamingResponseModel(token_streamer=token_streamer)

        return self.NBStreamingResponse(model_instance)



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.


class ChainError(BaseException):
    r"""Base error for any chain related errors."""
    pass


class ChainConnectionError(ChainError):
    r"""Error for any chain connection related errors."""
    pass


class ChainTransactionError(ChainError):
    r"""Error for any chain transaction related errors."""
    pass


class ChainQueryError(ChainError):
    r"""Error for any chain query related errors."""
    pass


class StakeError(ChainTransactionError):
    r"""Error raised when a stake transaction fails."""
    pass


class UnstakeError(ChainTransactionError):
    r"""Error raised when an unstake transaction fails."""
    pass


class IdentityError(ChainTransactionError):
    r"""Error raised when an identity transaction fails."""
    pass


class NominationError(ChainTransactionError):
    r"""Error raised when a nomination transaction fails."""
    pass


class TransferError(ChainTransactionError):
    r"""Error raised when a transfer transaction fails."""
    pass


class RegistrationError(ChainTransactionError):
    r"""Error raised when a particle registration transaction fails."""
    pass


class NotRegisteredError(ChainTransactionError):
    r"""Error raised when a particle is not registered, and the transaction requires it to be."""
    pass


class NotDelegateError(StakeError):
    r"""Error raised when a hotkey you are trying to stake to is not a delegate."""
    pass


class KeyFileError(Exception):
    """Error thrown when the keyfile is corrupt, non-writable, non-readable or the password used to decrypt is invalid."""

    pass


class MetadataError(ChainTransactionError):
    r"""Error raised when metadata commitment transaction fails."""
    pass


class InvalidRequestNameError(Exception):
    r"""This exception is raised when the request name is invalid. Ususally indicates a broken URL."""
    pass


class UnknownNucleonError(Exception):
    r"""This exception is raised when the request name is not found in the Fermion's forward_fns dictionary."""
    pass


class NucleonParsingError(Exception):
    r"""This exception is raised when the request headers are unable to be parsed into the nucleon type."""
    pass


class NotVerifiedException(Exception):
    r"""This exception is raised when the request is not verified."""
    pass


class BlacklistedException(Exception):
    r"""This exception is raised when the request is blacklisted."""
    pass


class PriorityException(Exception):
    r"""This exception is raised when the request priority is not met."""
    pass


class PostProcessException(Exception):
    r"""This exception is raised when the response headers cannot be updated."""
    pass


class RunException(Exception):
    r"""This exception is raised when the requested function cannot be executed. Indicates a server error."""
    pass


class InternalServerError(Exception):
    r"""This exception is raised when the requested function fails on the server. Indicates a server error."""
    pass



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import sys
import nimble
from tqdm import tqdm
from rich.prompt import Confirm, Prompt
from nimble.utils.balance import Balance
from typing import List, Union, Optional, Tuple
from .utils import get_hotkey_wallets_for_wallet
from . import defaults

console = nimble.__console__


class UnStakeCommand:
    """
    Executes the 'remove' command to unstake NIM tokens from one or more hotkeys and transfer them back to the user's coldkey on the nimble network.
    This command is used to withdraw tokens previously staked to different hotkeys.

    Usage:
    Users can specify the amount to unstake, the hotkeys to unstake from (either by name or SS58 address),
    and whether to unstake from all hotkeys. The command checks for sufficient stake and inferences for confirmation before proceeding with the unstaking process.

    Optional arguments:
    - --all (bool): When set, unstakes all staked tokens from the specified hotkeys.
    - --amount (float): The amount of NIM tokens to unstake.
    - --hotkey_ss58address (str): The SS58 address of the hotkey to unstake from.
    - --max_stake (float): Sets the maximum amount of NIM to remain staked in each hotkey.
    - --hotkeys (list): Specifies hotkeys by name or SS58 address to unstake from.
    - --all_hotkeys (bool): When set, unstakes from all hotkeys associated with the wallet, excluding any specified in --hotkeys.

    The command asks for confirmation before executing the unstaking operation.

    Example usage:
    >>> nbcli stake remove --amount 100 --hotkeys hk1,hk2

    Note:
    This command is important for users who wish to reallocate their stakes or withdraw them from the network.
    It allows for flexible management of token stakes across different particles (hotkeys) on the network.
    """

    @classmethod
    def check_config(cls, config: "nimble.config"):
        if not config.is_set("wallet.name") and not config.no_predict:
            wallet_name = Prompt.ask(
                "Enter wallet name", default=defaults.wallet.name
            )
            config.wallet.name = str(wallet_name)

        if (
            not config.get("hotkey_ss58address", d=None)
            and not config.is_set("wallet.hotkey")
            and not config.no_predict
            and not config.get("all_hotkeys")
            and not config.get("hotkeys")
        ):
            hotkey = Prompt.ask(
                "Enter hotkey name", default=defaults.wallet.hotkey
            )
            config.wallet.hotkey = str(hotkey)

        # Get amount.
        if (
            not config.get("hotkey_ss58address")
            and not config.get("amount")
            and not config.get("unstake_all")
            and not config.get("max_stake")
        ):
            hotkeys: str = ""
            if config.get("all_hotkeys"):
                hotkeys = "all hotkeys"
            elif config.get("hotkeys"):
                hotkeys = str(config.hotkeys).replace("[", "").replace("]", "")
            else:
                hotkeys = str(config.wallet.hotkey)
            if not Confirm.ask(
                "Unstake all NIM from: [bold]'{}'[/bold]?".format(hotkeys)
            ):
                amount = Prompt.ask("Enter NIM amount to unstake")
                config.unstake_all = False
                try:
                    config.amount = float(amount)
                except ValueError:
                    console.print(
                        ":cross_mark:[red] Invalid NIM amount[/red] [bold white]{}[/bold white]".format(
                            amount
                        )
                    )
                    sys.exit()
            else:
                config.unstake_all = True

    @staticmethod
    def add_args(command_parser):
        unstake_parser = command_parser.add_parser(
            "remove",
            help="""Remove stake from your hotkey accounts into their coldkey accounts.""",
        )
        unstake_parser.add_argument(
            "--all", dest="unstake_all", action="store_true", default=False
        )
        unstake_parser.add_argument(
            "--amount", dest="amount", type=float, required=False
        )
        unstake_parser.add_argument(
            "--hotkey_ss58address",
            dest="hotkey_ss58address",
            type=str,
            required=False,
        )
        unstake_parser.add_argument(
            "--max_stake",
            dest="max_stake",
            type=float,
            required=False,
            action="store",
            default=None,
            help="""Specify the maximum amount of NIM to have staked in each hotkey.""",
        )
        unstake_parser.add_argument(
            "--hotkeys",
            "--exclude_hotkeys",
            "--wallet.hotkeys",
            "--wallet.exclude_hotkeys",
            required=False,
            action="store",
            default=[],
            type=str,
            nargs="*",
            help="""Specify the hotkeys by name or ss58 address. (e.g. hk1 hk2 hk3)""",
        )
        unstake_parser.add_argument(
            "--all_hotkeys",
            "--wallet.all_hotkeys",
            required=False,
            action="store_true",
            default=False,
            help="""To specify all hotkeys. Specifying hotkeys will exclude them from this all.""",
        )
        nimble.wallet.add_args(unstake_parser)
        nimble.nbnetwork.add_args(unstake_parser)

    @staticmethod
    def run(cli):
        r"""Unstake token of amount from hotkey(s)."""
        config = cli.config.copy()
        wallet = nimble.wallet(config=config)
        nbnetwork: nimble.nbnetwork = nimble.nbnetwork(
            config=cli.config, log_verbose=False
        )

        # Get the hotkey_names (if any) and the hotkey_ss58s.
        hotkeys_to_unstake_from: List[Tuple[Optional[str], str]] = []
        if cli.config.get("hotkey_ss58address"):
            # Stake to specific hotkey.
            hotkeys_to_unstake_from = [
                (None, cli.config.get("hotkey_ss58address"))
            ]
        elif cli.config.get("all_hotkeys"):
            # Stake to all hotkeys.
            all_hotkeys: List[nimble.wallet] = get_hotkey_wallets_for_wallet(
                wallet=wallet
            )
            # Get the hotkeys to exclude. (d)efault to no exclusions.
            hotkeys_to_exclude: List[str] = cli.config.get("hotkeys", d=[])
            # Exclude hotkeys that are specified.
            hotkeys_to_unstake_from = [
                (wallet.hotkey_str, wallet.hotkey.ss58_address)
                for wallet in all_hotkeys
                if wallet.hotkey_str not in hotkeys_to_exclude
            ]  # definitely wallets

        elif cli.config.get("hotkeys"):
            # Stake to specific hotkeys.
            for hotkey_ss58_or_hotkey_name in cli.config.get("hotkeys"):
                if nimble.utils.is_valid_ss58_address(
                    hotkey_ss58_or_hotkey_name
                ):
                    # If the hotkey is a valid ss58 address, we add it to the list.
                    hotkeys_to_unstake_from.append(
                        (None, hotkey_ss58_or_hotkey_name)
                    )
                else:
                    # If the hotkey is not a valid ss58 address, we assume it is a hotkey name.
                    #  We then get the hotkey from the wallet and add it to the list.
                    wallet_ = nimble.wallet(
                        config=cli.config, hotkey=hotkey_ss58_or_hotkey_name
                    )
                    hotkeys_to_unstake_from.append(
                        (wallet_.hotkey_str, wallet_.hotkey.ss58_address)
                    )
        elif cli.config.wallet.get("hotkey"):
            # Only cli.config.wallet.hotkey is specified.
            #  so we stake to that single hotkey.
            hotkey_ss58_or_name = cli.config.wallet.get("hotkey")
            if nimble.utils.is_valid_ss58_address(hotkey_ss58_or_name):
                hotkeys_to_unstake_from = [(None, hotkey_ss58_or_name)]
            else:
                # Hotkey is not a valid ss58 address, so we assume it is a hotkey name.
                wallet_ = nimble.wallet(
                    config=cli.config, hotkey=hotkey_ss58_or_name
                )
                hotkeys_to_unstake_from = [
                    (wallet_.hotkey_str, wallet_.hotkey.ss58_address)
                ]
        else:
            # Only cli.config.wallet.hotkey is specified.
            #  so we stake to that single hotkey.
            assert cli.config.wallet.hotkey is not None
            hotkeys_to_unstake_from = [
                (None, nimble.wallet(config=cli.config).hotkey.ss58_address)
            ]

        final_hotkeys: List[Tuple[str, str]] = []
        final_amounts: List[Union[float, Balance]] = []
        for hotkey in tqdm(hotkeys_to_unstake_from):
            hotkey: Tuple[
                Optional[str], str
            ]  # (hotkey_name (or None), hotkey_ss58)
            unstake_amount_nim: float = cli.config.get(
                "amount"
            )  # The amount specified to unstake.
            hotkey_stake: Balance = nbnetwork.get_stake_for_coldkey_and_hotkey(
                hotkey_ss58=hotkey[1],
                coldkey_ss58=wallet.coldkeypub.ss58_address,
            )
            if unstake_amount_nim == None:
                unstake_amount_nim = hotkey_stake.nim
            if cli.config.get("max_stake"):
                # Get the current stake of the hotkey from this coldkey.
                unstake_amount_nim: float = hotkey_stake.nim - cli.config.get(
                    "max_stake"
                )
                cli.config.amount = unstake_amount_nim
                if unstake_amount_nim < 0:
                    # Skip if max_stake is greater than current stake.
                    continue
            else:
                if unstake_amount_nim is not None:
                    # There is a specified amount to unstake.
                    if unstake_amount_nim > hotkey_stake.nim:
                        # Skip if the specified amount is greater than the current stake.
                        continue

            final_amounts.append(unstake_amount_nim)
            final_hotkeys.append(
                hotkey
            )  # add both the name and the ss58 address.

        if len(final_hotkeys) == 0:
            # No hotkeys to unstake from.
            nimble.__console__.print(
                "Not enough stake to unstake from any hotkeys or max_stake is more than current stake."
            )
            return None

        # Ask to unstake
        if not cli.config.no_predict:
            if not Confirm.ask(
                f"Do you want to unstake from the following keys to {wallet.name}:\n"
                + "".join(
                    [
                        f"    [bold white]- {hotkey[0] + ':' if hotkey[0] else ''}{hotkey[1]}: {f'{amount} {nimble.__nim_symbol__}' if amount else 'All'}[/bold white]\n"
                        for hotkey, amount in zip(final_hotkeys, final_amounts)
                    ]
                )
            ):
                return None

        if len(final_hotkeys) == 1:
            # do regular unstake
            return nbnetwork.unstake(
                wallet=wallet,
                hotkey_ss58=final_hotkeys[0][1],
                amount=None
                if cli.config.get("unstake_all")
                else final_amounts[0],
                wait_for_inclusion=True,
                predict=not cli.config.no_predict,
            )

        nbnetwork.unstake_multiple(
            wallet=wallet,
            hotkey_ss58s=[hotkey_ss58 for _, hotkey_ss58 in final_hotkeys],
            amounts=None if cli.config.get("unstake_all") else final_amounts,
            wait_for_inclusion=True,
            predict=False,
        )



""" Implementation of the wallet class, which manages balances with staking and transfer. Also manages hotkey and coldkey.
"""
# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import os
import copy
import argparse
import nimble
from termcolor import colored
from substrateinterface import Keypair
from typing import Optional, Union, List, Tuple, Dict, overload
from nimble.utils import is_valid_nimble_address_or_public_key


def display_mnemonic_msg(keypair: Keypair, key_type: str):
    """
    Display the mnemonic and a warning message to keep the mnemonic safe.

    Args:
        keypair (Keypair): Keypair object.
        key_type (str): Type of the key (coldkey or hotkey).
    """
    mnemonic = keypair.mnemonic
    mnemonic_green = colored(mnemonic, "green")
    print(
        colored(
            "\nIMPORTANT: Store this mnemonic in a secure (preferable offline place), as anyone "
            "who has possession of this mnemonic can use it to regenerate the key and access your tokens. \n",
            "red",
        )
    )
    print(
        "The mnemonic to the new {} is:\n\n{}\n".format(
            key_type, mnemonic_green
        )
    )
    print(
        "You can use the mnemonic to recreate the key in case it gets lost. The command to use to regenerate the key using this mnemonic is:"
    )
    print("nbcli w regen_{} --mnemonic {}".format(key_type, mnemonic))
    print("")


class wallet:
    """
    The wallet class in the nimble framework handles wallet functionality, crucial for participating in
    the nimble network. It manages two types of keys: coldkey and hotkey, each serving different purposes
    in network operations. Each wallet contains a coldkey and a hotkey.

    The coldkey is the user's primary key for holding stake in their wallet and is the only way that users
    can access Nim. Coldkeys can hold tokens and should be encrypted on your device.

    The coldkey is the primary key used for securing the wallet's stake in the nimble network (Nim) and
    is critical for financial transactions like staking and unstaking tokens. It's recommended to keep the
    coldkey encrypted and secure, as it holds the actual tokens.

    The hotkey, in contrast, is used for operational tasks like subscribing to and setting weights in the
    network. It's linked to the coldkey through the megastring and does not directly hold tokens, thereby
    offering a safer way to interact with the network during regular operations.

    Attributes:
        name (str): The name of the wallet, used to identify it among possibly multiple wallets.
        path (str): File system path where wallet keys are stored.
        hotkey_str (str): String identifier for the hotkey.
        _hotkey, _coldkey, _coldkeypub (nimble.Keypair): Internal representations of the hotkey and coldkey.

    Methods:
        create_if_non_existent, create, recreate: Methods to handle the creation of wallet keys.
        get_coldkey, get_hotkey, get_coldkeypub: Methods to retrieve specific keys.
        set_coldkey, set_hotkey, set_coldkeypub: Methods to set or update keys.
        hotkey_file, coldkey_file, coldkeypub_file: Properties that return respective key file objects.
        regenerate_coldkey, regenerate_hotkey, regenerate_coldkeypub: Methods to regenerate keys from different sources.
        config, help, add_args: Utility methods for configuration and assistance.

    The wallet class is a fundamental component for users to interact securely with the nimble network,
    facilitating both operational tasks and transactions involving value transfer across the network.

    Example Usage:
        # Create a new wallet with default coldkey and hotkey names
        my_wallet = wallet()

        # Access hotkey and coldkey
        hotkey = my_wallet.get_hotkey()
        coldkey = my_wallet.get_coldkey()

        # Set a new coldkey
        my_wallet.new_coldkey(n_words=24) # number of seed words to use

        # Update wallet hotkey
        my_wallet.set_hotkey(new_hotkey)

        # Print wallet details
        print(my_wallet)

        # Access coldkey property, must use password to unlock
        my_wallet.coldkey
    """

    @classmethod
    def config(cls) -> "nimble.config":
        """
        Get config from the argument parser.

        Returns:
            nimble.config: Config object.
        """
        parser = argparse.ArgumentParser()
        cls.add_args(parser)
        return nimble.config(parser, args=[])

    @classmethod
    def help(cls):
        """
        Print help to stdout.
        """
        parser = argparse.ArgumentParser()
        cls.add_args(parser)
        print(cls.__new__.__doc__)
        parser.print_help()

    @classmethod
    def add_args(cls, parser: argparse.ArgumentParser, prefix: str = None):
        """
        Accept specific arguments from parser.

        Args:
            parser (argparse.ArgumentParser): Argument parser object.
            prefix (str): Argument prefix.
        """
        prefix_str = "" if prefix is None else prefix + "."
        try:
            default_name = os.getenv("NB_WALLET_NAME") or "default"
            default_hotkey = os.getenv("NB_WALLET_NAME") or "default"
            default_path = os.getenv("NB_WALLET_PATH") or "~/.nimble/wallets/"
            parser.add_argument(
                "--" + prefix_str + "wallet.name",
                required=False,
                default=default_name,
                help="The name of the wallet to unlock for running nimble "
                "(name mock is reserved for mocking this wallet)",
            )
            parser.add_argument(
                "--" + prefix_str + "wallet.hotkey",
                required=False,
                default=default_hotkey,
                help="The name of the wallet's hotkey.",
            )
            parser.add_argument(
                "--" + prefix_str + "wallet.path",
                required=False,
                default=default_path,
                help="The path to your nimble wallets",
            )
        except argparse.ArgumentError as e:
            pass

    def __init__(
        self,
        name: str = None,
        hotkey: str = None,
        path: str = None,
        config: "nimble.config" = None,
    ):
        r"""
        Initialize the nimble wallet object containing a hot and coldkey.

        Args:
            name (str, optional): The name of the wallet to unlock for running nimble. Defaults to 'default'.
            hotkey (str, optional): The name of hotkey used to running the miner. Defaults to 'default'.
            path (str, optional): The path to your nimble wallets. Defaults to '~/.nimble/wallets/'.
            config (nimble.config, optional): nimble.wallet.config(). Defaults to None.
        """
        # Fill config from passed args using command line defaults.
        if config is None:
            config = wallet.config()
        self.config = copy.deepcopy(config)
        self.config.wallet.name = name or self.config.wallet.get(
            "name", nimble.defaults.wallet.name
        )
        self.config.wallet.hotkey = hotkey or self.config.wallet.get(
            "hotkey", nimble.defaults.wallet.hotkey
        )
        self.config.wallet.path = path or self.config.wallet.get(
            "path", nimble.defaults.wallet.path
        )

        self.name = self.config.wallet.name
        self.path = self.config.wallet.path
        self.hotkey_str = self.config.wallet.hotkey

        self._hotkey = None
        self._coldkey = None
        self._coldkeypub = None

    def __str__(self):
        """
        Returns the string representation of the Wallet object.

        Returns:
            str: The string representation.
        """
        return "wallet({}, {}, {})".format(
            self.name, self.hotkey_str, self.path
        )

    def __repr__(self):
        """
        Returns the string representation of the Wallet object.

        Returns:
            str: The string representation.
        """
        return self.__str__()

    def create_if_non_existent(
        self,
        coldkey_use_password: bool = True,
        hotkey_use_password: bool = False,
    ) -> "wallet":
        """
        Checks for existing coldkeypub and hotkeys and creates them if non-existent.

        Args:
            coldkey_use_password (bool, optional): Whether to use a password for coldkey. Defaults to True.
            hotkey_use_password (bool, optional): Whether to use a password for hotkey. Defaults to False.

        Returns:
            wallet: The Wallet object.
        """
        return self.create(coldkey_use_password, hotkey_use_password)

    def create(
        self,
        coldkey_use_password: bool = True,
        hotkey_use_password: bool = False,
    ) -> "wallet":
        """
        Checks for existing coldkeypub and hotkeys and creates them if non-existent.

        Args:
            coldkey_use_password (bool, optional): Whether to use a password for coldkey. Defaults to True.
            hotkey_use_password (bool, optional): Whether to use a password for hotkey. Defaults to False.

        Returns:
            wallet: The Wallet object.
        """
        # ---- Setup Wallet. ----
        if (
            not self.coldkey_file.exists_on_device()
            and not self.coldkeypub_file.exists_on_device()
        ):
            self.create_new_coldkey(
                n_words=12, use_password=coldkey_use_password
            )

        if not self.hotkey_file.exists_on_device():
            self.create_new_hotkey(n_words=12, use_password=hotkey_use_password)

        return self

    def recreate(
        self,
        coldkey_use_password: bool = True,
        hotkey_use_password: bool = False,
    ) -> "wallet":
        """
        Checks for existing coldkeypub and hotkeys and creates them if non-existent.

        Args:
            coldkey_use_password (bool, optional): Whether to use a password for coldkey. Defaults to True.
            hotkey_use_password (bool, optional): Whether to use a password for hotkey. Defaults to False.

        Returns:
            wallet: The Wallet object.
        """
        # ---- Setup Wallet. ----
        self.create_new_coldkey(n_words=12, use_password=coldkey_use_password)
        self.create_new_hotkey(n_words=12, use_password=hotkey_use_password)
        return self

    @property
    def hotkey_file(self) -> "nimble.keyfile":
        """
        Property that returns the hotkey file.

        Returns:
            nimble.keyfile: The hotkey file.
        """
        wallet_path = os.path.expanduser(os.path.join(self.path, self.name))
        hotkey_path = os.path.join(wallet_path, "hotkeys", self.hotkey_str)
        return nimble.keyfile(path=hotkey_path)

    @property
    def coldkey_file(self) -> "nimble.keyfile":
        """
        Property that returns the coldkey file.

        Returns:
            nimble.keyfile: The coldkey file.
        """
        wallet_path = os.path.expanduser(os.path.join(self.path, self.name))
        coldkey_path = os.path.join(wallet_path, "coldkey")
        return nimble.keyfile(path=coldkey_path)

    @property
    def coldkeypub_file(self) -> "nimble.keyfile":
        """
        Property that returns the coldkeypub file.

        Returns:
            nimble.keyfile: The coldkeypub file.
        """
        wallet_path = os.path.expanduser(os.path.join(self.path, self.name))
        coldkeypub_path = os.path.join(wallet_path, "coldkeypub.txt")
        return nimble.keyfile(path=coldkeypub_path)

    def set_hotkey(
        self,
        keypair: "nimble.Keypair",
        encrypt: bool = False,
        overwrite: bool = False,
    ) -> "nimble.keyfile":
        """
        Sets the hotkey for the wallet.

        Args:
            keypair (nimble.Keypair): The hotkey keypair.
            encrypt (bool, optional): Whether to encrypt the hotkey. Defaults to False.
            overwrite (bool, optional): Whether to overwrite an existing hotkey. Defaults to False.

        Returns:
            nimble.keyfile: The hotkey file.
        """
        self._hotkey = keypair
        self.hotkey_file.set_keypair(
            keypair, encrypt=encrypt, overwrite=overwrite
        )

    def set_coldkeypub(
        self,
        keypair: "nimble.Keypair",
        encrypt: bool = False,
        overwrite: bool = False,
    ) -> "nimble.keyfile":
        """
        Sets the coldkeypub for the wallet.

        Args:
            keypair (nimble.Keypair): The coldkeypub keypair.
            encrypt (bool, optional): Whether to encrypt the coldkeypub. Defaults to False.
            overwrite (bool, optional): Whether to overwrite an existing coldkeypub. Defaults to False.

        Returns:
            nimble.keyfile: The coldkeypub file.
        """
        self._coldkeypub = nimble.Keypair(ss58_address=keypair.ss58_address)
        self.coldkeypub_file.set_keypair(
            self._coldkeypub, encrypt=encrypt, overwrite=overwrite
        )

    def set_coldkey(
        self,
        keypair: "nimble.Keypair",
        encrypt: bool = True,
        overwrite: bool = False,
    ) -> "nimble.keyfile":
        """
        Sets the coldkey for the wallet.

        Args:
            keypair (nimble.Keypair): The coldkey keypair.
            encrypt (bool, optional): Whether to encrypt the coldkey. Defaults to True.
            overwrite (bool, optional): Whether to overwrite an existing coldkey. Defaults to False.

        Returns:
            nimble.keyfile: The coldkey file.
        """
        self._coldkey = keypair
        self.coldkey_file.set_keypair(
            self._coldkey, encrypt=encrypt, overwrite=overwrite
        )

    def get_coldkey(self, password: str = None) -> "nimble.Keypair":
        """
        Gets the coldkey from the wallet.

        Args:
            password (str, optional): The password to decrypt the coldkey. Defaults to None.

        Returns:
            nimble.Keypair: The coldkey keypair.
        """
        return self.coldkey_file.get_keypair(password=password)

    def get_hotkey(self, password: str = None) -> "nimble.Keypair":
        """
        Gets the hotkey from the wallet.

        Args:
            password (str, optional): The password to decrypt the hotkey. Defaults to None.

        Returns:
            nimble.Keypair: The hotkey keypair.
        """
        return self.hotkey_file.get_keypair(password=password)

    def get_coldkeypub(self, password: str = None) -> "nimble.Keypair":
        """
        Gets the coldkeypub from the wallet.

        Args:
            password (str, optional): The password to decrypt the coldkeypub. Defaults to None.

        Returns:
            nimble.Keypair: The coldkeypub keypair.
        """
        return self.coldkeypub_file.get_keypair(password=password)

    @property
    def hotkey(self) -> "nimble.Keypair":
        r"""Loads the hotkey from wallet.path/wallet.name/hotkeys/wallet.hotkey or raises an error.
        Returns:
            hotkey (Keypair):
                hotkey loaded from config arguments.
        Raises:
            KeyFileError: Raised if the file is corrupt of non-existent.
            CryptoKeyError: Raised if the user enters an incorrec password for an encrypted keyfile.
        """
        if self._hotkey is None:
            self._hotkey = self.hotkey_file.keypair
        return self._hotkey

    @property
    def coldkey(self) -> "nimble.Keypair":
        r"""Loads the hotkey from wallet.path/wallet.name/coldkey or raises an error.
        Returns:
            coldkey (Keypair):
                colkey loaded from config arguments.
        Raises:
            KeyFileError: Raised if the file is corrupt of non-existent.
            CryptoKeyError: Raised if the user enters an incorrec password for an encrypted keyfile.
        """
        if self._coldkey is None:
            self._coldkey = self.coldkey_file.keypair
        return self._coldkey

    @property
    def coldkeypub(self) -> "nimble.Keypair":
        r"""Loads the coldkeypub from wallet.path/wallet.name/coldkeypub.txt or raises an error.
        Returns:
            coldkeypub (Keypair):
                colkeypub loaded from config arguments.
        Raises:
            KeyFileError: Raised if the file is corrupt of non-existent.
            CryptoKeyError: Raised if the user enters an incorrect password for an encrypted keyfile.
        """
        if self._coldkeypub is None:
            self._coldkeypub = self.coldkeypub_file.keypair
        return self._coldkeypub

    def create_coldkey_from_uri(
        self,
        uri: str,
        use_password: bool = True,
        overwrite: bool = False,
        suppress: bool = False,
    ) -> "wallet":
        """Creates coldkey from suri string, optionally encrypts it with the user's inputed password.
        Args:
            uri: (str, required):
                URI string to use i.e. /Alice or /Bob
            use_password (bool, optional):
                Is the created key password protected.
            overwrite (bool, optional):
                Will this operation overwrite the coldkey under the same path <wallet path>/<wallet name>/coldkey
        Returns:
            wallet (nimble.wallet):
                this object with newly created coldkey.
        """
        keypair = Keypair.create_from_uri(uri)
        if not suppress:
            display_mnemonic_msg(keypair, "coldkey")
        self.set_coldkey(keypair, encrypt=use_password, overwrite=overwrite)
        self.set_coldkeypub(keypair, overwrite=overwrite)
        return self

    def create_hotkey_from_uri(
        self,
        uri: str,
        use_password: bool = False,
        overwrite: bool = False,
        suppress: bool = False,
    ) -> "wallet":
        """Creates hotkey from suri string, optionally encrypts it with the user's inputed password.
        Args:
            uri: (str, required):
                URI string to use i.e. /Alice or /Bob
            use_password (bool, optional):
                Is the created key password protected.
            overwrite (bool, optional):
                Will this operation overwrite the hotkey under the same path <wallet path>/<wallet name>/hotkeys/<hotkey>
        Returns:
            wallet (nimble.wallet):
                this object with newly created hotkey.
        """
        keypair = Keypair.create_from_uri(uri)
        if not suppress:
            display_mnemonic_msg(keypair, "hotkey")
        self.set_hotkey(keypair, encrypt=use_password, overwrite=overwrite)
        return self

    def new_coldkey(
        self,
        n_words: int = 12,
        use_password: bool = True,
        overwrite: bool = False,
        suppress: bool = False,
    ) -> "wallet":
        """Creates a new coldkey, optionally encrypts it with the user's inputed password and saves to disk.
        Args:
            n_words: (int, optional):
                Number of mnemonic words to use.
            use_password (bool, optional):
                Is the created key password protected.
            overwrite (bool, optional):
                Will this operation overwrite the coldkey under the same path <wallet path>/<wallet name>/coldkey
        Returns:
            wallet (nimble.wallet):
                this object with newly created coldkey.
        """
        self.create_new_coldkey(n_words, use_password, overwrite, suppress)

    def create_new_coldkey(
        self,
        n_words: int = 12,
        use_password: bool = True,
        overwrite: bool = False,
        suppress: bool = False,
    ) -> "wallet":
        """Creates a new coldkey, optionally encrypts it with the user's inputed password and saves to disk.
        Args:
            n_words: (int, optional):
                Number of mnemonic words to use.
            use_password (bool, optional):
                Is the created key password protected.
            overwrite (bool, optional):
                Will this operation overwrite the coldkey under the same path <wallet path>/<wallet name>/coldkey
        Returns:
            wallet (nimble.wallet):
                this object with newly created coldkey.
        """
        mnemonic = Keypair.generate_mnemonic(n_words)
        keypair = Keypair.create_from_mnemonic(mnemonic)
        if not suppress:
            display_mnemonic_msg(keypair, "coldkey")
        self.set_coldkey(keypair, encrypt=use_password, overwrite=overwrite)
        self.set_coldkeypub(keypair, overwrite=overwrite)
        return self

    def new_hotkey(
        self,
        n_words: int = 12,
        use_password: bool = False,
        overwrite: bool = False,
        suppress: bool = False,
    ) -> "wallet":
        """Creates a new hotkey, optionally encrypts it with the user's inputed password and saves to disk.
        Args:
            n_words: (int, optional):
                Number of mnemonic words to use.
            use_password (bool, optional):
                Is the created key password protected.
            overwrite (bool, optional):
                Will this operation overwrite the hotkey under the same path <wallet path>/<wallet name>/hotkeys/<hotkey>
        Returns:
            wallet (nimble.wallet):
                this object with newly created hotkey.
        """
        self.create_new_hotkey(n_words, use_password, overwrite, suppress)

    def create_new_hotkey(
        self,
        n_words: int = 12,
        use_password: bool = False,
        overwrite: bool = False,
        suppress: bool = False,
    ) -> "wallet":
        """Creates a new hotkey, optionally encrypts it with the user's inputed password and saves to disk.
        Args:
            n_words: (int, optional):
                Number of mnemonic words to use.
            use_password (bool, optional):
                Is the created key password protected.
            overwrite (bool, optional):
                Will this operation overwrite the hotkey under the same path <wallet path>/<wallet name>/hotkeys/<hotkey>
        Returns:
            wallet (nimble.wallet):
                this object with newly created hotkey.
        """
        mnemonic = Keypair.generate_mnemonic(n_words)
        keypair = Keypair.create_from_mnemonic(mnemonic)
        if not suppress:
            display_mnemonic_msg(keypair, "hotkey")
        self.set_hotkey(keypair, encrypt=use_password, overwrite=overwrite)
        return self

    def regenerate_coldkeypub(
        self,
        ss58_address: Optional[str] = None,
        public_key: Optional[Union[str, bytes]] = None,
        overwrite: bool = False,
        suppress: bool = False,
    ) -> "wallet":
        """Regenerates the coldkeypub from passed ss58_address or public_key and saves the file
           Requires either ss58_address or public_key to be passed.
        Args:
            ss58_address: (str, optional):
                Address as ss58 string.
            public_key: (str | bytes, optional):
                Public key as hex string or bytes.
            overwrite (bool, optional) (default: False):
                Will this operation overwrite the coldkeypub (if exists) under the same path <wallet path>/<wallet name>/coldkeypub
        Returns:
            wallet (nimble.wallet):
                newly re-generated Wallet with coldkeypub.

        """
        if ss58_address is None and public_key is None:
            raise ValueError("Either ss58_address or public_key must be passed")

        if not is_valid_nimble_address_or_public_key(
            ss58_address if ss58_address is not None else public_key
        ):
            raise ValueError(
                f"Invalid {'ss58_address' if ss58_address is not None else 'public_key'}"
            )

        if ss58_address is not None:
            ss58_format = nimble.utils.get_ss58_format(ss58_address)
            keypair = Keypair(
                ss58_address=ss58_address,
                public_key=public_key,
                ss58_format=ss58_format,
            )
        else:
            keypair = Keypair(
                ss58_address=ss58_address,
                public_key=public_key,
                ss58_format=nimble.__ss58_format__,
            )

        # No need to encrypt the public key
        self.set_coldkeypub(keypair, overwrite=overwrite)

        return self

    # Short name for regenerate_coldkeypub
    regen_coldkeypub = regenerate_coldkeypub

    @overload
    def regenerate_coldkey(
        self,
        mnemonic: Optional[Union[list, str]] = None,
        use_password: bool = True,
        overwrite: bool = False,
        suppress: bool = False,
    ) -> "wallet":
        ...

    @overload
    def regenerate_coldkey(
        self,
        seed: Optional[str] = None,
        use_password: bool = True,
        overwrite: bool = False,
        suppress: bool = False,
    ) -> "wallet":
        ...

    @overload
    def regenerate_coldkey(
        self,
        json: Optional[Tuple[Union[str, Dict], str]] = None,
        use_password: bool = True,
        overwrite: bool = False,
        suppress: bool = False,
    ) -> "wallet":
        ...

    def regenerate_coldkey(
        self,
        use_password: bool = True,
        overwrite: bool = False,
        suppress: bool = False,
        **kwargs,
    ) -> "wallet":
        """Regenerates the coldkey from passed mnemonic, seed, or json encrypts it with the user's password and saves the file
        Args:
            mnemonic: (Union[list, str], optional):
                Key mnemonic as list of words or string space separated words.
            seed: (str, optional):
                Seed as hex string.
            json: (Tuple[Union[str, Dict], str], optional):
                Restore from encrypted JSON backup as (json_data: Union[str, Dict], passphrase: str)
            use_password (bool, optional):
                Is the created key password protected.
            overwrite (bool, optional):
                Will this operation overwrite the coldkey under the same path <wallet path>/<wallet name>/coldkey
        Returns:
            wallet (nimble.wallet):
                this object with newly created coldkey.

        Note: uses priority order: mnemonic > seed > json
        """
        if len(kwargs) == 0:
            raise ValueError("Must pass either mnemonic, seed, or json")

        # Get from kwargs
        mnemonic = kwargs.get("mnemonic", None)
        seed = kwargs.get("seed", None)
        json = kwargs.get("json", None)

        if mnemonic is None and seed is None and json is None:
            raise ValueError("Must pass either mnemonic, seed, or json")

        if mnemonic is not None:
            if isinstance(mnemonic, str):
                mnemonic = mnemonic.split()
            if len(mnemonic) not in [12, 15, 18, 21, 24]:
                raise ValueError(
                    "Mnemonic has invalid size. This should be 12,15,18,21 or 24 words"
                )
            keypair = Keypair.create_from_mnemonic(
                " ".join(mnemonic), ss58_format=nimble.__ss58_format__
            )
            if not suppress:
                display_mnemonic_msg(keypair, "coldkey")
        elif seed is not None:
            keypair = Keypair.create_from_seed(
                seed, ss58_format=nimble.__ss58_format__
            )
        else:
            # json is not None
            if (
                not isinstance(json, tuple)
                or len(json) != 2
                or not isinstance(json[0], (str, dict))
                or not isinstance(json[1], str)
            ):
                raise ValueError(
                    "json must be a tuple of (json_data: str | Dict, passphrase: str)"
                )

            json_data, passphrase = json
            keypair = Keypair.create_from_encrypted_json(
                json_data, passphrase, ss58_format=nimble.__ss58_format__
            )

        self.set_coldkey(keypair, encrypt=use_password, overwrite=overwrite)
        self.set_coldkeypub(keypair, overwrite=overwrite)
        return self

    # Short name for regenerate_coldkey
    regen_coldkey = regenerate_coldkey

    @overload
    def regenerate_hotkey(
        self,
        mnemonic: Optional[Union[list, str]] = None,
        use_password: bool = True,
        overwrite: bool = False,
        suppress: bool = False,
    ) -> "wallet":
        ...

    @overload
    def regenerate_hotkey(
        self,
        seed: Optional[str] = None,
        use_password: bool = True,
        overwrite: bool = False,
        suppress: bool = False,
    ) -> "wallet":
        ...

    @overload
    def regenerate_hotkey(
        self,
        json: Optional[Tuple[Union[str, Dict], str]] = None,
        use_password: bool = True,
        overwrite: bool = False,
        suppress: bool = False,
    ) -> "wallet":
        ...

    def regenerate_hotkey(
        self,
        use_password: bool = True,
        overwrite: bool = False,
        suppress: bool = False,
        **kwargs,
    ) -> "wallet":
        """Regenerates the hotkey from passed mnemonic, encrypts it with the user's password and save the file
        Args:
            mnemonic: (Union[list, str], optional):
                Key mnemonic as list of words or string space separated words.
            seed: (str, optional):
                Seed as hex string.
            json: (Tuple[Union[str, Dict], str], optional):
                Restore from encrypted JSON backup as (json_data: Union[str, Dict], passphrase: str)
            use_password (bool, optional):
                Is the created key password protected.
            overwrite (bool, optional):
                Will this operation overwrite the hotkey under the same path <wallet path>/<wallet name>/hotkeys/<hotkey>
        Returns:
            wallet (nimble.wallet):
                this object with newly created hotkey.
        """
        if len(kwargs) == 0:
            raise ValueError("Must pass either mnemonic, seed, or json")

        # Get from kwargs
        mnemonic = kwargs.get("mnemonic", None)
        seed = kwargs.get("seed", None)
        json = kwargs.get("json", None)

        if mnemonic is None and seed is None and json is None:
            raise ValueError("Must pass either mnemonic, seed, or json")

        if mnemonic is not None:
            if isinstance(mnemonic, str):
                mnemonic = mnemonic.split()
            if len(mnemonic) not in [12, 15, 18, 21, 24]:
                raise ValueError(
                    "Mnemonic has invalid size. This should be 12,15,18,21 or 24 words"
                )
            keypair = Keypair.create_from_mnemonic(
                " ".join(mnemonic), ss58_format=nimble.__ss58_format__
            )
            if not suppress:
                display_mnemonic_msg(keypair, "hotkey")
        elif seed is not None:
            keypair = Keypair.create_from_seed(
                seed, ss58_format=nimble.__ss58_format__
            )
        else:
            # json is not None
            if (
                not isinstance(json, tuple)
                or len(json) != 2
                or not isinstance(json[0], (str, dict))
                or not isinstance(json[1], str)
            ):
                raise ValueError(
                    "json must be a tuple of (json_data: str | Dict, passphrase: str)"
                )

            json_data, passphrase = json
            keypair = Keypair.create_from_encrypted_json(
                json_data, passphrase, ss58_format=nimble.__ss58_format__
            )

        self.set_hotkey(keypair, encrypt=use_password, overwrite=overwrite)
        return self

    # Short name for regenerate_hotkey
    regen_hotkey = regenerate_hotkey



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import math


def get_human_readable(num, suffix="H"):
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1000.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1000.0
    return f"{num:.1f}Y{suffix}"


def millify(n: int):
    millnames = ["", " K", " M", " B", " T"]
    n = float(n)
    millidx = max(
        0,
        min(
            len(millnames) - 1,
            int(math.floor(0 if n == 0 else math.log10(abs(n)) / 3)),
        ),
    )

    return "{:.2f}{}".format(n / 10 ** (3 * millidx), millnames[millidx])



# Copyright 2009 Brian Quinlan. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.

"""Implements ThreadPoolExecutor."""

__author__ = "Brian Quinlan (brian@sweetapp.com)"

import argparse
import itertools
import nimble
import os
import queue
import random
import sys
import threading
import time
import weakref

from concurrent.futures import _base
from loguru import logger
from typing import Callable

# Workers are created as daemon threads. This is done to allow the interpreter
# to exit when there are still idle threads in a ThreadPoolExecutor's thread
# pool (i.e. shutdown() was not called). However, allowing workers to die with
# the interpreter has two undesirable properties:
#   - The workers would still be running during interpreter shutdown,
#     meaning that they would fail in unpredictable ways.
#   - The workers could be killed while evaluating a work item, which could
#     be bad if the callable being evaluated has external side-effects e.g.
#     writing to a file.
#
# To work around this problem, an exit handler is installed which tells the
# workers to exit when their work queues are empty and then waits until the
# threads finish.
# Ref: https://github.com/python/cpython/blob/main/Lib/concurrent/futures/thread.py

_threads_queues = weakref.WeakKeyDictionary()
_shutdown = False


class _WorkItem(object):
    def __init__(self, future, fn, start_time, args, kwargs):
        self.future = future
        self.fn = fn
        self.start_time = start_time
        self.args = args
        self.kwargs = kwargs

    def run(self):
        """Run the given work item"""
        # Checks if future is canceled or if work item is stale
        if (not self.future.set_running_or_notify_cancel()) or (
            time.time() - self.start_time > nimble.__blocktime__
        ):
            return

        try:
            result = self.fn(*self.args, **self.kwargs)
        except BaseException as exc:
            self.future.set_exception(exc)
            # Break a reference cycle with the exception 'exc'
            self = None
        else:
            self.future.set_result(result)


NULL_ENTRY = (sys.maxsize, _WorkItem(None, None, time.time(), (), {}))


def _worker(executor_reference, work_queue, initializer, initargs):
    if initializer is not None:
        try:
            initializer(*initargs)
        except BaseException:
            _base.LOGGER.critical("Exception in initializer:", exc_info=True)
            executor = executor_reference()
            if executor is not None:
                executor._initializer_failed()
            return
    try:
        while True:
            work_item = work_queue.get(block=True)
            priority = work_item[0]
            item = work_item[1]
            if priority == sys.maxsize:
                del item
            elif item is not None:
                item.run()
                # Delete references to object. See issue16284
                del item
                continue

            executor = executor_reference()
            # Exit if:
            #   - The interpreter is shutting down OR
            #   - The executor that owns the worker has been collected OR
            #   - The executor that owns the worker has been shutdown.
            if _shutdown or executor is None or executor._shutdown:
                # Flag the executor as shutting down as early as possible if it
                # is not gc-ed yet.
                if executor is not None:
                    executor._shutdown = True
                # Notice other workers
                work_queue.put(NULL_ENTRY)
                return
            del executor
    except BaseException:
        logger.error("work_item", work_item)
        _base.LOGGER.critical("Exception in worker", exc_info=True)


class BrokenThreadPool(_base.BrokenExecutor):
    """
    Raised when a worker thread in a ThreadPoolExecutor failed initializing.
    """


class PriorityThreadPoolExecutor(_base.Executor):
    """Base threadpool executor with a priority queue"""

    # Used to assign unique thread names when thread_name_prefix is not supplied.
    _counter = itertools.count().__next__

    def __init__(
        self,
        maxsize=-1,
        max_workers=None,
        thread_name_prefix="",
        initializer=None,
        initargs=(),
    ):
        """Initializes a new ThreadPoolExecutor instance.
        Args:
            max_workers: The maximum number of threads that can be used to
                execute the given calls.
            thread_name_prefix: An optional name prefix to give our threads.
            initializer: An callable used to initialize worker threads.
            initargs: A tuple of arguments to pass to the initializer.
        """
        if max_workers is None:
            # Use this number because ThreadPoolExecutor is often
            # used to overlap I/O instead of CPU work.
            max_workers = (os.cpu_count() or 1) * 5
        if max_workers <= 0:
            raise ValueError("max_workers must be greater than 0")

        if initializer is not None and not callable(initializer):
            raise TypeError("initializer must be a callable")

        self._max_workers = max_workers
        self._work_queue = queue.PriorityQueue(maxsize=maxsize)
        self._idle_semaphore = threading.Semaphore(0)
        self._threads = set()
        self._broken = False
        self._shutdown = False
        self._shutdown_lock = threading.Lock()
        self._thread_name_prefix = thread_name_prefix or (
            "ThreadPoolExecutor-%d" % self._counter()
        )
        self._initializer = initializer
        self._initargs = initargs

    @classmethod
    def add_args(cls, parser: argparse.ArgumentParser, prefix: str = None):
        """Accept specific arguments from parser"""
        prefix_str = "" if prefix == None else prefix + "."
        try:
            default_max_workers = (
                os.getenv("NB_PRIORITY_MAX_WORKERS")
                if os.getenv("NB_PRIORITY_MAX_WORKERS") != None
                else 5
            )
            default_maxsize = (
                os.getenv("NB_PRIORITY_MAXSIZE")
                if os.getenv("NB_PRIORITY_MAXSIZE") != None
                else 10
            )
            parser.add_argument(
                "--" + prefix_str + "priority.max_workers",
                type=int,
                help="""maximum number of threads in thread pool""",
                default=default_max_workers,
            )
            parser.add_argument(
                "--" + prefix_str + "priority.maxsize",
                type=int,
                help="""maximum size of tasks in priority queue""",
                default=default_maxsize,
            )
        except argparse.ArgumentError:
            # re-parsing arguments.
            pass

    @classmethod
    def config(cls) -> "nimble.config":
        """Get config from the argument parser
        Return: nimble.config object
        """
        parser = argparse.ArgumentParser()
        PriorityThreadPoolExecutor.add_args(parser)
        return nimble.config(parser, args=[])

    @property
    def is_empty(self):
        return self._work_queue.empty()

    def submit(self, fn: Callable, *args, **kwargs) -> _base.Future:
        with self._shutdown_lock:
            if self._broken:
                raise BrokenThreadPool(self._broken)

            if self._shutdown:
                raise RuntimeError("cannot schedule new futures after shutdown")
            if _shutdown:
                raise RuntimeError(
                    "cannot schedule new futures after " "interpreter shutdown"
                )

            priority = kwargs.get("priority", random.randint(0, 1000000))
            if priority == 0:
                priority = random.randint(1, 100)
            eplison = random.uniform(0, 0.01) * priority
            start_time = time.time()
            if "priority" in kwargs:
                del kwargs["priority"]

            f = _base.Future()
            w = _WorkItem(f, fn, start_time, args, kwargs)
            self._work_queue.put((-float(priority + eplison), w), block=False)
            self._adjust_thread_count()
            return f

    submit.__doc__ = _base.Executor.submit.__doc__

    def _adjust_thread_count(self):
        # if idle threads are available, don't spin new threads
        if self._idle_semaphore.acquire(timeout=0):
            return

        # When the executor gets lost, the weakref callback will wake up
        # the worker threads.
        def weakref_cb(_, q=self._work_queue):
            q.put(NULL_ENTRY)

        num_threads = len(self._threads)
        if num_threads < self._max_workers:
            thread_name = "%s_%d" % (
                self._thread_name_prefix or self,
                num_threads,
            )
            t = threading.Thread(
                name=thread_name,
                target=_worker,
                args=(
                    weakref.ref(self, weakref_cb),
                    self._work_queue,
                    self._initializer,
                    self._initargs,
                ),
            )
            t.daemon = True
            t.start()
            self._threads.add(t)
            _threads_queues[t] = self._work_queue

    def _initializer_failed(self):
        with self._shutdown_lock:
            self._broken = (
                "A thread initializer failed, the thread pool "
                "is not usable anymore"
            )
            # Drain work queue and mark pending futures failed
            while True:
                try:
                    work_item = self._work_queue.get_nowait()
                except queue.Empty:
                    break
                if work_item is not None:
                    work_item.future.set_exception(
                        BrokenThreadPool(self._broken)
                    )

    def shutdown(self, wait=True):
        with self._shutdown_lock:
            self._shutdown = True
            self._work_queue.put(NULL_ENTRY)

        if wait:
            for t in self._threads:
                try:
                    t.join(timeout=2)
                except Exception:
                    pass

    shutdown.__doc__ = _base.Executor.shutdown.__doc__



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import base64
import msgpack
import msgpack_numpy
import numpy
import pydantic
import torch

from nimble.constants import TORCH_DTYPES
from typing import Optional, Union, List


def cast_dtype(raw: Union[None, torch.dtype, str]) -> str:
    """
    Casts the raw value to a string representing the torch data type.

    Args:
        raw (Union[None, torch.dtype, str]): The raw value to cast.

    Returns:
        str: The string representing the torch data type.

    Raises:
        Exception: If the raw value is of an invalid type.
    """
    if not raw:
        return None

    if isinstance(raw, torch.dtype):
        return TORCH_DTYPES[raw]

    if isinstance(raw, str):
        assert (
            raw in TORCH_DTYPES
        ), f"{str} not a valid torch type in dict {TORCH_DTYPES}"
        return raw

    raise Exception(
        f"{raw} of type {type(raw)} does not have a valid type in Union[None, torch.dtype, str]"
    )


def cast_shape(raw: Union[None, List[int], str]) -> str:
    """
    Casts the raw value to a string representing the tensor shape.

    Args:
        raw (Union[None, List[int], str]): The raw value to cast.

    Returns:
        str: The string representing the tensor shape.

    Raises:
        Exception: If the raw value is of an invalid type or if the list elements are not of type int.
    """
    if not raw:
        return None

    if isinstance(raw, list):
        if len(raw) == 0:
            return raw
        elif isinstance(raw[0], int):
            return raw
        else:
            raise Exception(f"{raw} list elements are not of type int")

    if isinstance(raw, str):
        shape = list(map(int, raw.split("[")[1].split("]")[0].split(",")))
        return shape

    raise Exception(
        f"{raw} of type {type(raw)} does not have a valid type in Union[None, List[int], str]"
    )


class tensor:
    def __new__(cls, tensor: Union[list, numpy.ndarray, torch.Tensor]):
        if isinstance(tensor, list) or isinstance(tensor, numpy.ndarray):
            tensor = torch.tensor(tensor)

        return Tensor.serialize(tensor=tensor)


class Tensor(pydantic.BaseModel):
    """
    Represents a Tensor object.

    Attributes:
        buffer (Optional[str]): Tensor buffer data.
        dtype (str): Tensor data type.
        shape (List[int]): Tensor shape.
    """

    class Config:
        validate_assignment = True

    def tensor(self) -> torch.Tensor:
        return self.deserialize()

    def tolist(self) -> List[object]:
        return self.deserialize().tolist()

    def numpy(self) -> "numpy.ndarray":
        return self.deserialize().detach().numpy()

    def deserialize(self) -> "torch.Tensor":
        """
        Deserializes the Tensor object.

        Returns:
            torch.Tensor: The deserialized tensor object.

        Raises:
            Exception: If the deserialization process encounters an error.
        """
        shape = tuple(self.shape)
        buffer_bytes = base64.b64decode(self.buffer.encode("utf-8"))
        numpy_object = msgpack.unpackb(
            buffer_bytes, object_hook=msgpack_numpy.decode
        ).copy()
        torch_object = torch.as_tensor(numpy_object)
        # Reshape does not work for (0) or [0]
        if not (len(shape) == 1 and shape[0] == 0):
            torch_object = torch_object.reshape(shape)
        return torch_object.type(TORCH_DTYPES[self.dtype])

    @staticmethod
    def serialize(tensor: "torch.Tensor") -> "Tensor":
        """
        Serializes the given tensor.

        Args:
            tensor (torch.Tensor): The tensor to serialize.

        Returns:
            Tensor: The serialized tensor.

        Raises:
            Exception: If the serialization process encounters an error.
        """
        dtype = str(tensor.dtype)
        shape = list(tensor.shape)
        if len(shape) == 0:
            shape = [0]
        torch_numpy = tensor.cpu().detach().numpy().copy()
        data_buffer = base64.b64encode(
            msgpack.packb(torch_numpy, default=msgpack_numpy.encode)
        ).decode("utf-8")
        return Tensor(buffer=data_buffer, shape=shape, dtype=dtype)

    buffer: Optional[str] = pydantic.Field(
        title="buffer",
        description="Tensor buffer data. This field stores the serialized representation of the tensor data.",
        examples="0x321e13edqwds231231231232131",
        allow_mutation=False,
        repr=False,
    )  # Represents the tensor buffer data.

    dtype: str = pydantic.Field(
        title="dtype",
        description="Tensor data type. This field specifies the data type of the tensor, such as torch.float32 or torch.int64.",
        examples="torch.float32",
        allow_mutation=False,
        repr=True,
    )  # Represents the data type of the tensor.
    _extract_dtype = pydantic.validator("dtype", pre=True, allow_reuse=True)(
        cast_dtype
    )

    shape: List[int] = pydantic.Field(
        title="shape",
        description="Tensor shape. This field defines the dimensions of the tensor as a list of integers, such as [10, 10] for a 2D tensor with shape (10, 10).",
        examples="[10,10]",
        allow_mutation=False,
        repr=True,
    )  # Represents the shape of the tensor.
    _extract_shape = pydantic.validator("shape", pre=True, allow_reuse=True)(
        cast_shape
    )



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from substrateinterface.utils import ss58
from typing import Union, Optional

from .. import __ss58_format__
from substrateinterface import Keypair as Keypair


def get_ss58_format(ss58_address: str) -> int:
    """Returns the ss58 format of the given ss58 address."""
    return ss58.get_ss58_format(ss58_address)


def is_valid_ss58_address(address: str) -> bool:
    """
    Checks if the given address is a valid ss58 address.

    Args:
        address(str): The address to check.

    Returns:
        True if the address is a valid ss58 address for nimble, False otherwise.
    """
    try:
        return ss58.is_valid_ss58_address(
            address, valid_ss58_format=__ss58_format__
        ) or ss58.is_valid_ss58_address(
            address, valid_ss58_format=42
        )  # Default substrate ss58 format (legacy)
    except IndexError:
        return False


def is_valid_ed25519_pubkey(public_key: Union[str, bytes]) -> bool:
    """
    Checks if the given public_key is a valid ed25519 key.

    Args:
        public_key(Union[str, bytes]): The public_key to check.

    Returns:
        True if the public_key is a valid ed25519 key, False otherwise.

    """
    try:
        if isinstance(public_key, str):
            if len(public_key) != 64 and len(public_key) != 66:
                raise ValueError("a public_key should be 64 or 66 characters")
        elif isinstance(public_key, bytes):
            if len(public_key) != 32:
                raise ValueError("a public_key should be 32 bytes")
        else:
            raise ValueError("public_key must be a string or bytes")

        keypair = Keypair(public_key=public_key, ss58_format=__ss58_format__)

        ss58_addr = keypair.ss58_address
        return ss58_addr is not None

    except (ValueError, IndexError):
        return False


def is_valid_nimble_address_or_public_key(address: Union[str, bytes]) -> bool:
    """
    Checks if the given address is a valid destination address.

    Args:
        address(Union[str, bytes]): The address to check.

    Returns:
        True if the address is a valid destination address, False otherwise.
    """
    if isinstance(address, str):
        # Check if ed25519
        if address.startswith("0x"):
            return is_valid_ed25519_pubkey(address)
        else:
            # Assume ss58 address
            return is_valid_ss58_address(address)
    elif isinstance(address, bytes):
        # Check if ed25519
        return is_valid_ed25519_pubkey(address)
    else:
        # Invalid address type
        return False


def create_identity_dict(
    display: str = "",
    legal: str = "",
    web: str = "",
    riot: str = "",
    email: str = "",
    pgp_fingerprint: Optional[str] = None,
    image: str = "",
    info: str = "",
    twitter: str = "",
) -> dict:
    """
    Creates a dictionary with structure for identity extrinsic. Must fit within 64 bits.

    Args:
    display (str): String to be converted and stored under 'display'.
    legal (str): String to be converted and stored under 'legal'.
    web (str): String to be converted and stored under 'web'.
    riot (str): String to be converted and stored under 'riot'.
    email (str): String to be converted and stored under 'email'.
    pgp_fingerprint (str): String to be converted and stored under 'pgp_fingerprint'.
    image (str): String to be converted and stored under 'image'.
    info (str): String to be converted and stored under 'info'.
    twitter (str): String to be converted and stored under 'twitter'.

    Returns:
    dict: A dictionary with the specified structure and byte string conversions.

    Raises:
    ValueError: If pgp_fingerprint is not exactly 20 bytes long when encoded.
    """
    if pgp_fingerprint and len(pgp_fingerprint.encode()) != 20:
        raise ValueError(
            "pgp_fingerprint must be exactly 20 bytes long when encoded"
        )

    return {
        "info": {
            "additional": [[]],
            "display": {f"Raw{len(display.encode())}": display.encode()},
            "legal": {f"Raw{len(legal.encode())}": legal.encode()},
            "web": {f"Raw{len(web.encode())}": web.encode()},
            "riot": {f"Raw{len(riot.encode())}": riot.encode()},
            "email": {f"Raw{len(email.encode())}": email.encode()},
            "pgp_fingerprint": pgp_fingerprint.encode()
            if pgp_fingerprint
            else None,
            "image": {f"Raw{len(image.encode())}": image.encode()},
            "info": {f"Raw{len(info.encode())}": info.encode()},
            "twitter": {f"Raw{len(twitter.encode())}": twitter.encode()},
        }
    }


def decode_hex_identity_dict(info_dictionary):
    for key, value in info_dictionary.items():
        if isinstance(value, dict):
            item = list(value.values())[0]
            if isinstance(item, str) and item.startswith("0x"):
                try:
                    info_dictionary[key] = bytes.fromhex(item[2:]).decode()
                except UnicodeDecodeError:
                    print(f"Could not decode: {key}: {item}")
            else:
                info_dictionary[key] = item
    return info_dictionary



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from random import randint
from types import SimpleNamespace
from typing import Any, Dict, List, Optional, Tuple, TypedDict, Union
from unittest.mock import MagicMock
from dataclasses import dataclass
from abc import abstractclassmethod
from collections.abc import Mapping

from hashlib import sha256
from ..wallet import wallet

from ..chain_data import (
    ParticleInfo,
    ParticleInfoLite,
    ProtonInfo,
    DelegateInfo,
    CosmosInfo,
    FermionInfo,
)
from ..errors import *
from ..nimble_network import NimbleNetwork
from ..utils import VIMPERNIM, U16_NORMALIZED_FLOAT
from ..utils.balance import Balance
from ..utils.registration import POWSolution

from typing import TypedDict


# Mock Testing Constant
__GLOBAL_MOCK_STATE__ = {}


class FermionServeCallParams(TypedDict):
    """
    Fermion serve chain call parameters.
    """

    version: int
    ip: int
    port: int
    ip_type: int
    netuid: int


class ProtonServeCallParams(TypedDict):
    """
    Proton serve chain call parameters.
    """

    version: int
    ip: int
    port: int
    ip_type: int
    netuid: int


BlockNumber = int


class InfoDict(Mapping):
    @abstractclassmethod
    def default(cls):
        raise NotImplementedError

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)


@dataclass
class FermionInfoDict(InfoDict):
    block: int
    version: int
    ip: int  # integer representation of ip address
    port: int
    ip_type: int
    protocol: int
    placeholder1: int  # placeholder for future use
    placeholder2: int

    @classmethod
    def default(cls):
        return cls(
            block=0,
            version=0,
            ip=0,
            port=0,
            ip_type=0,
            protocol=0,
            placeholder1=0,
            placeholder2=0,
        )


@dataclass
class ProtonInfoDict(InfoDict):
    block: int
    version: int
    ip: int  # integer representation of ip address
    port: int
    ip_type: int

    @classmethod
    def default(cls):
        return cls(block=0, version=0, ip=0, port=0, ip_type=0)


@dataclass
class MockNbNetworkValue:
    value: Optional[Any]


class MockMapResult:
    records: Optional[List[Tuple[MockNbNetworkValue, MockNbNetworkValue]]]

    def __init__(
        self,
        records: Optional[
            List[
                Tuple[
                    Union[Any, MockNbNetworkValue],
                    Union[Any, MockNbNetworkValue],
                ]
            ]
        ] = None,
    ):
        _records = [
            (
                MockNbNetworkValue(value=record[0]),
                MockNbNetworkValue(value=record[1]),
            )
            # Make sure record is a tuple of MockNbNetworkValue (dict with value attr)
            if not (
                isinstance(record, tuple)
                and all(
                    isinstance(item, dict) and hasattr(item, "value")
                    for item in record
                )
            )
            else record
            for record in records
        ]

        self.records = _records

    def __iter__(self):
        return iter(self.records)


class MockSystemState(TypedDict):
    Account: Dict[str, Dict[int, int]]  # address -> block -> balance


class MockNbNetworkState(TypedDict):
    Rho: Dict[int, Dict[BlockNumber, int]]  # netuid -> block -> rho
    Kappa: Dict[int, Dict[BlockNumber, int]]  # netuid -> block -> kappa
    Difficulty: Dict[
        int, Dict[BlockNumber, int]
    ]  # netuid -> block -> difficulty
    ImmunityPeriod: Dict[
        int, Dict[BlockNumber, int]
    ]  # netuid -> block -> immunity_period
    ValidatorBatchSize: Dict[
        int, Dict[BlockNumber, int]
    ]  # netuid -> block -> validator_batch_size
    Active: Dict[int, Dict[BlockNumber, bool]]  # (netuid, uid), block -> active
    Stake: Dict[
        str, Dict[str, Dict[int, int]]
    ]  # (hotkey, coldkey) -> block -> stake

    Delegates: Dict[str, Dict[int, float]]  # address -> block -> delegate_take

    NetworksAdded: Dict[
        int, Dict[BlockNumber, bool]
    ]  # netuid -> block -> added


class MockChainState(TypedDict):
    System: MockSystemState
    NbNetworkModule: MockNbNetworkState


class MockNbNetwork(NimbleNetwork):
    """
    A Mock NbNetwork class for running tests.
    This should mock only methods that make queries to the chain.
    e.g. We mock `NbNetwork.query_nbnetwork` instead of all query methods.

    This class will also store a local (mock) state of the chain.
    """

    chain_state: MockChainState
    block_number: int

    @classmethod
    def reset(cls) -> None:
        __GLOBAL_MOCK_STATE__.clear()

        _ = cls()

    def setup(self) -> None:
        if (
            not hasattr(self, "chain_state")
            or getattr(self, "chain_state") is None
        ):
            self.chain_state = {
                "System": {"Account": {}},
                "Balances": {"ExistentialDeposit": {0: 500}},
                "NbNetworkModule": {
                    "NetworksAdded": {},
                    "Rho": {},
                    "Kappa": {},
                    "Difficulty": {},
                    "ImmunityPeriod": {},
                    "ValidatorBatchSize": {},
                    "ValidatorSequenceLength": {},
                    "ValidatorEpochsPerReset": {},
                    "ValidatorEpochLength": {},
                    "MaxAllowedValidators": {},
                    "MinAllowedWeights": {},
                    "MaxWeightLimit": {},
                    "SynergyScalingLawPower": {},
                    "ScalingLawPower": {},
                    "CosmosN": {},
                    "MaxAllowedUids": {},
                    "NetworkModality": {},
                    "BlocksSinceLastStep": {},
                    "Tempo": {},
                    "NetworkConnect": {},
                    "EmissionValues": {},
                    "Burn": {},
                    "Active": {},
                    "Uids": {},
                    "Keys": {},
                    "Owner": {},
                    "IsNetworkMember": {},
                    "LastUpdate": {},
                    "Rank": {},
                    "Emission": {},
                    "Incentive": {},
                    "Consensus": {},
                    "Trust": {},
                    "ValidatorTrust": {},
                    "Dividends": {},
                    "PruningScores": {},
                    "ValidatorPermit": {},
                    "Weights": {},
                    "Bonds": {},
                    "Stake": {},
                    "TotalStake": {0: 0},
                    "TotalIssuance": {0: 0},
                    "TotalHotkeyStake": {},
                    "TotalColdkeyStake": {},
                    "TxRateLimit": {0: 0},  # No limit
                    "Delegates": {},
                    "Fermions": {},
                    "Proton": {},
                    "CosmosOwner": {},
                },
            }

            self.block_number = 0

            self.network = "mock"
            self.chain_endpoint = "mock_endpoint"
            self.substrate = MagicMock()

    def __init__(self, *args, **kwargs) -> None:
        self.__dict__ = __GLOBAL_MOCK_STATE__

        if (
            not hasattr(self, "chain_state")
            or getattr(self, "chain_state") is None
        ):
            self.setup()

    def get_block_hash(self, block_id: int) -> str:
        return "0x" + sha256(str(block_id).encode()).hexdigest()[:64]

    def create_cosmos(self, netuid: int) -> None:
        nbnetwork_state = self.chain_state["NbNetworkModule"]
        if netuid not in nbnetwork_state["NetworksAdded"]:
            # Per Cosmos
            nbnetwork_state["Rho"][netuid] = {}
            nbnetwork_state["Rho"][netuid][0] = 10
            nbnetwork_state["Kappa"][netuid] = {}
            nbnetwork_state["Kappa"][netuid][0] = 32_767
            nbnetwork_state["Difficulty"][netuid] = {}
            nbnetwork_state["Difficulty"][netuid][0] = 10_000_000
            nbnetwork_state["ImmunityPeriod"][netuid] = {}
            nbnetwork_state["ImmunityPeriod"][netuid][0] = 4096
            nbnetwork_state["ValidatorBatchSize"][netuid] = {}
            nbnetwork_state["ValidatorBatchSize"][netuid][0] = 32
            nbnetwork_state["ValidatorSequenceLength"][netuid] = {}
            nbnetwork_state["ValidatorSequenceLength"][netuid][0] = 256
            nbnetwork_state["ValidatorEpochsPerReset"][netuid] = {}
            nbnetwork_state["ValidatorEpochsPerReset"][netuid][0] = 60
            nbnetwork_state["ValidatorEpochLength"][netuid] = {}
            nbnetwork_state["ValidatorEpochLength"][netuid][0] = 100
            nbnetwork_state["MaxAllowedValidators"][netuid] = {}
            nbnetwork_state["MaxAllowedValidators"][netuid][0] = 128
            nbnetwork_state["MinAllowedWeights"][netuid] = {}
            nbnetwork_state["MinAllowedWeights"][netuid][0] = 1024
            nbnetwork_state["MaxWeightLimit"][netuid] = {}
            nbnetwork_state["MaxWeightLimit"][netuid][0] = 1_000
            nbnetwork_state["SynergyScalingLawPower"][netuid] = {}
            nbnetwork_state["SynergyScalingLawPower"][netuid][0] = 50
            nbnetwork_state["ScalingLawPower"][netuid] = {}
            nbnetwork_state["ScalingLawPower"][netuid][0] = 50
            nbnetwork_state["CosmosN"][netuid] = {}
            nbnetwork_state["CosmosN"][netuid][0] = 0
            nbnetwork_state["MaxAllowedUids"][netuid] = {}
            nbnetwork_state["MaxAllowedUids"][netuid][0] = 4096
            nbnetwork_state["NetworkModality"][netuid] = {}
            nbnetwork_state["NetworkModality"][netuid][0] = 0
            nbnetwork_state["BlocksSinceLastStep"][netuid] = {}
            nbnetwork_state["BlocksSinceLastStep"][netuid][0] = 0
            nbnetwork_state["Tempo"][netuid] = {}
            nbnetwork_state["Tempo"][netuid][0] = 99
            # nbnetwork_state['NetworkConnect'][netuid] = {}
            # nbnetwork_state['NetworkConnect'][netuid][0] = {}
            nbnetwork_state["EmissionValues"][netuid] = {}
            nbnetwork_state["EmissionValues"][netuid][0] = 0
            nbnetwork_state["Burn"][netuid] = {}
            nbnetwork_state["Burn"][netuid][0] = 0

            # Per-UID/Hotkey

            nbnetwork_state["Uids"][netuid] = {}
            nbnetwork_state["Keys"][netuid] = {}
            nbnetwork_state["Owner"][netuid] = {}

            nbnetwork_state["LastUpdate"][netuid] = {}
            nbnetwork_state["Active"][netuid] = {}
            nbnetwork_state["Rank"][netuid] = {}
            nbnetwork_state["Emission"][netuid] = {}
            nbnetwork_state["Incentive"][netuid] = {}
            nbnetwork_state["Consensus"][netuid] = {}
            nbnetwork_state["Trust"][netuid] = {}
            nbnetwork_state["ValidatorTrust"][netuid] = {}
            nbnetwork_state["Dividends"][netuid] = {}
            nbnetwork_state["PruningScores"][netuid] = {}
            nbnetwork_state["PruningScores"][netuid][0] = {}
            nbnetwork_state["ValidatorPermit"][netuid] = {}

            nbnetwork_state["Weights"][netuid] = {}
            nbnetwork_state["Bonds"][netuid] = {}

            nbnetwork_state["Fermions"][netuid] = {}
            nbnetwork_state["Proton"][netuid] = {}

            nbnetwork_state["NetworksAdded"][netuid] = {}
            nbnetwork_state["NetworksAdded"][netuid][0] = True

        else:
            raise Exception("Cosmos already exists")

    def set_difficulty(self, netuid: int, difficulty: int) -> None:
        nbnetwork_state = self.chain_state["NbNetworkModule"]
        if netuid not in nbnetwork_state["NetworksAdded"]:
            raise Exception("Cosmos does not exist")

        nbnetwork_state["Difficulty"][netuid][self.block_number] = difficulty

    def _register_particle(self, netuid: int, hotkey: str, coldkey: str) -> int:
        nbnetwork_state = self.chain_state["NbNetworkModule"]
        if netuid not in nbnetwork_state["NetworksAdded"]:
            raise Exception("Cosmos does not exist")

        cosmos_n = self._get_most_recent_storage(
            nbnetwork_state["CosmosN"][netuid]
        )

        if cosmos_n > 0 and any(
            self._get_most_recent_storage(nbnetwork_state["Keys"][netuid][uid])
            == hotkey
            for uid in range(cosmos_n)
        ):
            # already_registered
            raise Exception("Hotkey already registered")
        else:
            # Not found
            if cosmos_n >= self._get_most_recent_storage(
                nbnetwork_state["MaxAllowedUids"][netuid]
            ):
                # Cosmos full, replace particle randomly
                uid = randint(0, cosmos_n - 1)
            else:
                # Cosmos not full, add new particle
                # Append as next uid and increment cosmos_n
                uid = cosmos_n
                nbnetwork_state["CosmosN"][netuid][self.block_number] = (
                    cosmos_n + 1
                )

            nbnetwork_state["Stake"][hotkey] = {}
            nbnetwork_state["Stake"][hotkey][coldkey] = {}
            nbnetwork_state["Stake"][hotkey][coldkey][self.block_number] = 0

            nbnetwork_state["Uids"][netuid][hotkey] = {}
            nbnetwork_state["Uids"][netuid][hotkey][self.block_number] = uid

            nbnetwork_state["Keys"][netuid][uid] = {}
            nbnetwork_state["Keys"][netuid][uid][self.block_number] = hotkey

            nbnetwork_state["Owner"][hotkey] = {}
            nbnetwork_state["Owner"][hotkey][self.block_number] = coldkey

            nbnetwork_state["Active"][netuid][uid] = {}
            nbnetwork_state["Active"][netuid][uid][self.block_number] = True

            nbnetwork_state["LastUpdate"][netuid][uid] = {}
            nbnetwork_state["LastUpdate"][netuid][uid][
                self.block_number
            ] = self.block_number

            nbnetwork_state["Rank"][netuid][uid] = {}
            nbnetwork_state["Rank"][netuid][uid][self.block_number] = 0.0

            nbnetwork_state["Emission"][netuid][uid] = {}
            nbnetwork_state["Emission"][netuid][uid][self.block_number] = 0.0

            nbnetwork_state["Incentive"][netuid][uid] = {}
            nbnetwork_state["Incentive"][netuid][uid][self.block_number] = 0.0

            nbnetwork_state["Consensus"][netuid][uid] = {}
            nbnetwork_state["Consensus"][netuid][uid][self.block_number] = 0.0

            nbnetwork_state["Trust"][netuid][uid] = {}
            nbnetwork_state["Trust"][netuid][uid][self.block_number] = 0.0

            nbnetwork_state["ValidatorTrust"][netuid][uid] = {}
            nbnetwork_state["ValidatorTrust"][netuid][uid][
                self.block_number
            ] = 0.0

            nbnetwork_state["Dividends"][netuid][uid] = {}
            nbnetwork_state["Dividends"][netuid][uid][self.block_number] = 0.0

            nbnetwork_state["PruningScores"][netuid][uid] = {}
            nbnetwork_state["PruningScores"][netuid][uid][
                self.block_number
            ] = 0.0

            nbnetwork_state["ValidatorPermit"][netuid][uid] = {}
            nbnetwork_state["ValidatorPermit"][netuid][uid][
                self.block_number
            ] = False

            nbnetwork_state["Weights"][netuid][uid] = {}
            nbnetwork_state["Weights"][netuid][uid][self.block_number] = []

            nbnetwork_state["Bonds"][netuid][uid] = {}
            nbnetwork_state["Bonds"][netuid][uid][self.block_number] = []

            nbnetwork_state["Fermions"][netuid][hotkey] = {}
            nbnetwork_state["Fermions"][netuid][hotkey][self.block_number] = {}

            nbnetwork_state["Proton"][netuid][hotkey] = {}
            nbnetwork_state["Proton"][netuid][hotkey][
                self.block_number
            ] = {}

            if hotkey not in nbnetwork_state["IsNetworkMember"]:
                nbnetwork_state["IsNetworkMember"][hotkey] = {}
            nbnetwork_state["IsNetworkMember"][hotkey][netuid] = {}
            nbnetwork_state["IsNetworkMember"][hotkey][netuid][
                self.block_number
            ] = True

            return uid

    @staticmethod
    def _convert_to_balance(balance: Union["Balance", float, int]) -> "Balance":
        if isinstance(balance, float):
            balance = Balance.from_nim(balance)

        if isinstance(balance, int):
            balance = Balance.from_vim(balance)

        return balance

    def force_register_particle(
        self,
        netuid: int,
        hotkey: str,
        coldkey: str,
        stake: Union["Balance", float, int] = Balance(0),
        balance: Union["Balance", float, int] = Balance(0),
    ) -> int:
        """
        Force register a particle on the mock chain, returning the UID.
        """
        stake = self._convert_to_balance(stake)
        balance = self._convert_to_balance(balance)

        nbnetwork_state = self.chain_state["NbNetworkModule"]
        if netuid not in nbnetwork_state["NetworksAdded"]:
            raise Exception("Cosmos does not exist")

        uid = self._register_particle(
            netuid=netuid, hotkey=hotkey, coldkey=coldkey
        )

        nbnetwork_state["TotalStake"][self.block_number] = (
            self._get_most_recent_storage(nbnetwork_state["TotalStake"])
            + stake.vim
        )
        nbnetwork_state["Stake"][hotkey][coldkey][self.block_number] = stake.vim

        if balance.vim > 0:
            self.force_set_balance(coldkey, balance)
        self.force_set_balance(coldkey, balance)

        return uid

    def force_set_balance(
        self,
        ss58_address: str,
        balance: Union["Balance", float, int] = Balance(0),
    ) -> Tuple[bool, Optional[str]]:
        """
        Returns:
            Tuple[bool, Optional[str]]: (success, err_msg)
        """
        balance = self._convert_to_balance(balance)

        if ss58_address not in self.chain_state["System"]["Account"]:
            self.chain_state["System"]["Account"][ss58_address] = {
                "data": {"free": {0: 0}}
            }

        old_balance = self.get_balance(ss58_address, self.block_number)
        diff = balance.vim - old_balance.vim

        # Update total issuance
        self.chain_state["NbNetworkModule"]["TotalIssuance"][
            self.block_number
        ] = (
            self._get_most_recent_storage(
                self.chain_state["NbNetworkModule"]["TotalIssuance"]
            )
            + diff
        )

        self.chain_state["System"]["Account"][ss58_address] = {
            "data": {"free": {self.block_number: balance.vim}}
        }

        return True, None

    # Alias for force_set_balance
    sudo_force_set_balance = force_set_balance

    def do_block_step(self) -> None:
        self.block_number += 1

        # Doesn't do epoch
        nbnetwork_state = self.chain_state["NbNetworkModule"]
        for cosmos in nbnetwork_state["NetworksAdded"]:
            nbnetwork_state["BlocksSinceLastStep"][cosmos][
                self.block_number
            ] = (
                self._get_most_recent_storage(
                    nbnetwork_state["BlocksSinceLastStep"][cosmos]
                )
                + 1
            )

    def _handle_type_default(self, name: str, params: List[object]) -> object:
        defaults_mapping = {
            "TotalStake": 0,
            "TotalHotkeyStake": 0,
            "TotalColdkeyStake": 0,
            "Stake": 0,
        }

        return defaults_mapping.get(name, None)

    def query_nbnetwork(
        self,
        name: str,
        block: Optional[int] = None,
        params: Optional[List[object]] = [],
    ) -> MockNbNetworkValue:
        if block:
            if self.block_number < block:
                raise Exception("Cannot query block in the future")

        else:
            block = self.block_number

        state = self.chain_state["NbNetworkModule"][name]
        if state is not None:
            # Use prefix
            if len(params) > 0:
                while state is not None and len(params) > 0:
                    state = state.get(params.pop(0), None)
                    if state is None:
                        return SimpleNamespace(
                            value=self._handle_type_default(name, params)
                        )

            # Use block
            state_at_block = state.get(block, None)
            while state_at_block is None and block > 0:
                block -= 1
                state_at_block = self.state.get(block, None)
            if state_at_block is not None:
                return SimpleNamespace(value=state_at_block)

            return SimpleNamespace(
                value=self._handle_type_default(name, params)
            )
        else:
            return SimpleNamespace(
                value=self._handle_type_default(name, params)
            )

    def query_map_nbnetwork(
        self,
        name: str,
        block: Optional[int] = None,
        params: Optional[List[object]] = [],
    ) -> Optional[MockMapResult]:
        """
        Note: Double map requires one param
        """
        if block:
            if self.block_number < block:
                raise Exception("Cannot query block in the future")

        else:
            block = self.block_number

        state = self.chain_state["NbNetworkModule"][name]
        if state is not None:
            # Use prefix
            if len(params) > 0:
                while state is not None and len(params) > 0:
                    state = state.get(params.pop(0), None)
                    if state is None:
                        return MockMapResult([])

            # Check if single map or double map
            if len(state.keys()) == 0:
                return MockMapResult([])

            inner = list(state.values())[0]
            # Should have at least one key
            if len(inner.keys()) == 0:
                raise Exception("Invalid state")

            # Check if double map
            if isinstance(list(inner.values())[0], dict):
                # is double map
                raise ChainQueryError("Double map requires one param")

            # Iterate over each key and add value to list, max at block
            records = []
            for key in state:
                result = self._get_most_recent_storage(state[key], block)
                if result is None:
                    continue  # Skip if no result for this key at `block` or earlier

                records.append((key, result))

            return MockMapResult(records)
        else:
            return MockMapResult([])

    def query_constant(
        self, module_name: str, constant_name: str, block: Optional[int] = None
    ) -> Optional[object]:
        if block:
            if self.block_number < block:
                raise Exception("Cannot query block in the future")

        else:
            block = self.block_number

        state = self.chain_state.get(module_name, None)
        if state is not None:
            if constant_name in state:
                state = state[constant_name]
            else:
                return None

            # Use block
            state_at_block = self._get_most_recent_storage(state, block)
            if state_at_block is not None:
                return SimpleNamespace(value=state_at_block)

            return state_at_block["data"]["free"]  # Can be None
        else:
            return None

    def get_current_block(self) -> int:
        return self.block_number

    # ==== Balance RPC methods ====

    def get_balance(self, address: str, block: int = None) -> "Balance":
        if block:
            if self.block_number < block:
                raise Exception("Cannot query block in the future")

        else:
            block = self.block_number

        state = self.chain_state["System"]["Account"]
        if state is not None:
            if address in state:
                state = state[address]
            else:
                return Balance(0)

            # Use block
            balance_state = state["data"]["free"]
            state_at_block = self._get_most_recent_storage(
                balance_state, block
            )  # Can be None
            if state_at_block is not None:
                bal_as_int = state_at_block
                return Balance.from_vim(bal_as_int)
            else:
                return Balance(0)
        else:
            return Balance(0)

    def get_balances(self, block: int = None) -> Dict[str, "Balance"]:
        balances = {}
        for address in self.chain_state["System"]["Account"]:
            balances[address] = self.get_balance(address, block)

        return balances

    # ==== Particle RPC methods ====

    def particle_for_uid(
        self, uid: int, netuid: int, block: Optional[int] = None
    ) -> Optional[ParticleInfo]:
        if uid is None:
            return ParticleInfo._null_particle()

        if block:
            if self.block_number < block:
                raise Exception("Cannot query block in the future")

        else:
            block = self.block_number

        if netuid not in self.chain_state["NbNetworkModule"]["NetworksAdded"]:
            return None

        particle_info = self._particle_cosmos_exists(uid, netuid, block)
        if particle_info is None:
            return None

        else:
            return particle_info

    def particles(
        self, netuid: int, block: Optional[int] = None
    ) -> List[ParticleInfo]:
        if netuid not in self.chain_state["NbNetworkModule"]["NetworksAdded"]:
            raise Exception("Cosmos does not exist")

        particles = []
        cosmos_n = self._get_most_recent_storage(
            self.chain_state["NbNetworkModule"]["CosmosN"][netuid], block
        )
        for uid in range(cosmos_n):
            particle_info = self.particle_for_uid(uid, netuid, block)
            if particle_info is not None:
                particles.append(particle_info)

        return particles

    @staticmethod
    def _get_most_recent_storage(
        storage: Dict[BlockNumber, Any], block_number: Optional[int] = None
    ) -> Any:
        if block_number is None:
            items = list(storage.items())
            items.sort(key=lambda x: x[0], reverse=True)
            if len(items) == 0:
                return None

            return items[0][1]

        else:
            while block_number >= 0:
                if block_number in storage:
                    return storage[block_number]

                block_number -= 1

            return None

    def _get_fermion_info(
        self, netuid: int, hotkey: str, block: Optional[int] = None
    ) -> FermionInfoDict:
        # Fermions [netuid][hotkey][block_number]
        nbnetwork_state = self.chain_state["NbNetworkModule"]
        if netuid not in nbnetwork_state["Fermions"]:
            return FermionInfoDict.default()

        if hotkey not in nbnetwork_state["Fermions"][netuid]:
            return FermionInfoDict.default()

        result = self._get_most_recent_storage(
            nbnetwork_state["Fermions"][netuid][hotkey], block
        )
        if not result:
            return FermionInfoDict.default()

        return result

    def _get_proton_info(
        self, netuid: int, hotkey: str, block: Optional[int] = None
    ) -> ProtonInfoDict:
        nbnetwork_state = self.chain_state["NbNetworkModule"]
        if netuid not in nbnetwork_state["Proton"]:
            return ProtonInfoDict.default()

        if hotkey not in nbnetwork_state["Proton"][netuid]:
            return ProtonInfoDict.default()

        result = self._get_most_recent_storage(
            nbnetwork_state["Proton"][netuid][hotkey], block
        )
        if not result:
            return ProtonInfoDict.default()

        return result

    def _particle_cosmos_exists(
        self, uid: int, netuid: int, block: Optional[int] = None
    ) -> Optional[ParticleInfo]:
        nbnetwork_state = self.chain_state["NbNetworkModule"]
        if netuid not in nbnetwork_state["NetworksAdded"]:
            return None

        if (
            self._get_most_recent_storage(
                nbnetwork_state["CosmosN"][netuid]
            )
            <= uid
        ):
            return None

        hotkey = self._get_most_recent_storage(
            nbnetwork_state["Keys"][netuid][uid]
        )
        if hotkey is None:
            return None

        fermion_info_ = self._get_fermion_info(netuid, hotkey, block)

        proton_info = self._get_proton_info(netuid, hotkey, block)

        coldkey = self._get_most_recent_storage(
            nbnetwork_state["Owner"][hotkey], block
        )
        active = self._get_most_recent_storage(
            nbnetwork_state["Active"][netuid][uid], block
        )
        rank = self._get_most_recent_storage(
            nbnetwork_state["Rank"][netuid][uid], block
        )
        emission = self._get_most_recent_storage(
            nbnetwork_state["Emission"][netuid][uid], block
        )
        incentive = self._get_most_recent_storage(
            nbnetwork_state["Incentive"][netuid][uid], block
        )
        consensus = self._get_most_recent_storage(
            nbnetwork_state["Consensus"][netuid][uid], block
        )
        trust = self._get_most_recent_storage(
            nbnetwork_state["Trust"][netuid][uid], block
        )
        validator_trust = self._get_most_recent_storage(
            nbnetwork_state["ValidatorTrust"][netuid][uid], block
        )
        dividends = self._get_most_recent_storage(
            nbnetwork_state["Dividends"][netuid][uid], block
        )
        pruning_score = self._get_most_recent_storage(
            nbnetwork_state["PruningScores"][netuid][uid], block
        )
        last_update = self._get_most_recent_storage(
            nbnetwork_state["LastUpdate"][netuid][uid], block
        )
        validator_permit = self._get_most_recent_storage(
            nbnetwork_state["ValidatorPermit"][netuid][uid], block
        )

        weights = self._get_most_recent_storage(
            nbnetwork_state["Weights"][netuid][uid], block
        )
        bonds = self._get_most_recent_storage(
            nbnetwork_state["Bonds"][netuid][uid], block
        )

        stake_dict = {
            coldkey: Balance.from_vim(
                self._get_most_recent_storage(
                    nbnetwork_state["Stake"][hotkey][coldkey], block
                )
            )
            for coldkey in nbnetwork_state["Stake"][hotkey]
        }

        stake = sum(stake_dict.values())

        weights = [[int(weight[0]), int(weight[1])] for weight in weights]
        bonds = [[int(bond[0]), int(bond[1])] for bond in bonds]
        rank = U16_NORMALIZED_FLOAT(rank)
        emission = emission / VIMPERNIM
        incentive = U16_NORMALIZED_FLOAT(incentive)
        consensus = U16_NORMALIZED_FLOAT(consensus)
        trust = U16_NORMALIZED_FLOAT(trust)
        validator_trust = U16_NORMALIZED_FLOAT(validator_trust)
        dividends = U16_NORMALIZED_FLOAT(dividends)
        proton_info = ProtonInfo.fix_decoded_values(proton_info)
        fermion_info_ = FermionInfo.from_particle_info(
            {"hotkey": hotkey, "coldkey": coldkey, "fermion_info": fermion_info_}
        )

        particle_info = ParticleInfo(
            hotkey=hotkey,
            coldkey=coldkey,
            uid=uid,
            netuid=netuid,
            active=active,
            rank=rank,
            emission=emission,
            incentive=incentive,
            consensus=consensus,
            trust=trust,
            validator_trust=validator_trust,
            dividends=dividends,
            pruning_score=pruning_score,
            last_update=last_update,
            validator_permit=validator_permit,
            stake=stake,
            stake_dict=stake_dict,
            total_stake=stake,
            proton_info=proton_info,
            fermion_info=fermion_info_,
            weights=weights,
            bonds=bonds,
            is_null=False,
        )

        return particle_info

    def particle_for_uid_lite(
        self, uid: int, netuid: int, block: Optional[int] = None
    ) -> Optional[ParticleInfoLite]:
        if block:
            if self.block_number < block:
                raise Exception("Cannot query block in the future")

        else:
            block = self.block_number

        if netuid not in self.chain_state["NbNetworkModule"]["NetworksAdded"]:
            raise Exception("Cosmos does not exist")

        particle_info = self._particle_cosmos_exists(uid, netuid, block)
        if particle_info is None:
            return None

        else:
            particle_info_dict = particle_info.__dict__
            del particle_info
            del particle_info_dict["weights"]
            del particle_info_dict["bonds"]

            particle_info_lite = ParticleInfoLite(**particle_info_dict)
            return particle_info_lite

    def particles_lite(
        self, netuid: int, block: Optional[int] = None
    ) -> List[ParticleInfoLite]:
        if netuid not in self.chain_state["NbNetworkModule"]["NetworksAdded"]:
            raise Exception("Cosmos does not exist")

        particles = []
        cosmos_n = self._get_most_recent_storage(
            self.chain_state["NbNetworkModule"]["CosmosN"][netuid]
        )
        for uid in range(cosmos_n):
            particle_info = self.particle_for_uid_lite(uid, netuid, block)
            if particle_info is not None:
                particles.append(particle_info)

        return particles

    # Extrinsics
    def _do_delegation(
        self,
        wallet: "wallet",
        delegate_ss58: str,
        amount: "Balance",
        wait_for_inclusion: bool = True,
        wait_for_finalization: bool = False,
    ) -> bool:
        # Check if delegate
        if not self.is_hotkey_delegate(hotkey_ss58=delegate_ss58):
            raise Exception("Not a delegate")

        # do stake
        success = self._do_stake(
            wallet=wallet,
            hotkey_ss58=delegate_ss58,
            amount=amount,
            wait_for_inclusion=wait_for_inclusion,
            wait_for_finalization=wait_for_finalization,
        )

        return success

    def _do_undelegation(
        self,
        wallet: "wallet",
        delegate_ss58: str,
        amount: "Balance",
        wait_for_inclusion: bool = True,
        wait_for_finalization: bool = False,
    ) -> bool:
        # Check if delegate
        if not self.is_hotkey_delegate(hotkey_ss58=delegate_ss58):
            raise Exception("Not a delegate")

        # do unstake
        self._do_unstake(
            wallet=wallet,
            hotkey_ss58=delegate_ss58,
            amount=amount,
            wait_for_inclusion=wait_for_inclusion,
            wait_for_finalization=wait_for_finalization,
        )

    def _do_nominate(
        self,
        wallet: "wallet",
        wait_for_inclusion: bool = True,
        wait_for_finalization: bool = False,
    ) -> bool:
        hotkey_ss58 = wallet.hotkey.ss58_address
        coldkey_ss58 = wallet.coldkeypub.ss58_address

        nbnetwork_state = self.chain_state["NbNetworkModule"]
        if self.is_hotkey_delegate(hotkey_ss58=hotkey_ss58):
            return True

        else:
            nbnetwork_state["Delegates"][hotkey_ss58] = {}
            nbnetwork_state["Delegates"][hotkey_ss58][
                self.block_number
            ] = 0.18  # Constant for now

            return True

    def get_transfer_fee(
        self, wallet: "wallet", dest: str, value: Union["Balance", float, int]
    ) -> "Balance":
        return Balance(700)

    def _do_transfer(
        self,
        wallet: "wallet",
        dest: str,
        transfer_balance: "Balance",
        wait_for_inclusion: bool = True,
        wait_for_finalization: bool = False,
    ) -> Tuple[bool, Optional[str], Optional[str]]:
        bal = self.get_balance(wallet.coldkeypub.ss58_address)
        dest_bal = self.get_balance(dest)
        transfer_fee = self.get_transfer_fee(wallet, dest, transfer_balance)

        existential_deposit = self.get_existential_deposit()

        if bal < transfer_balance + existential_deposit + transfer_fee:
            raise Exception("Insufficient balance")

        # Remove from the free balance
        self.chain_state["System"]["Account"][wallet.coldkeypub.ss58_address][
            "data"
        ]["free"][self.block_number] = (
            bal - transfer_balance - transfer_fee
        ).vim

        # Add to the free balance
        if dest not in self.chain_state["System"]["Account"]:
            self.chain_state["System"]["Account"][dest] = {"data": {"free": {}}}

        self.chain_state["System"]["Account"][dest]["data"]["free"][
            self.block_number
        ] = (dest_bal + transfer_balance).vim

        return True, None, None

    def _do_pow_register(
        self,
        netuid: int,
        wallet: "wallet",
        pow_result: "POWSolution",
        wait_for_inclusion: bool = False,
        wait_for_finalization: bool = True,
    ) -> Tuple[bool, Optional[str]]:
        # Assume pow result is valid

        nbnetwork_state = self.chain_state["NbNetworkModule"]
        if netuid not in nbnetwork_state["NetworksAdded"]:
            raise Exception("Cosmos does not exist")

        self._register_particle(
            netuid=netuid,
            hotkey=wallet.hotkey.ss58_address,
            coldkey=wallet.coldkeypub.ss58_address,
        )

        return True, None

    def _do_burned_register(
        self,
        netuid: int,
        wallet: "wallet",
        wait_for_inclusion: bool = False,
        wait_for_finalization: bool = True,
    ) -> Tuple[bool, Optional[str]]:
        nbnetwork_state = self.chain_state["NbNetworkModule"]
        if netuid not in nbnetwork_state["NetworksAdded"]:
            raise Exception("Cosmos does not exist")

        bal = self.get_balance(wallet.coldkeypub.ss58_address)
        burn = self.burn(netuid=netuid)
        existential_deposit = self.get_existential_deposit()

        if bal < burn + existential_deposit:
            raise Exception("Insufficient funds")

        self._register_particle(
            netuid=netuid,
            hotkey=wallet.hotkey.ss58_address,
            coldkey=wallet.coldkeypub.ss58_address,
        )

        # Burn the funds
        self.chain_state["System"]["Account"][wallet.coldkeypub.ss58_address][
            "data"
        ]["free"][self.block_number] = (bal - burn).vim

        return True, None

    def _do_stake(
        self,
        wallet: "wallet",
        hotkey_ss58: str,
        amount: "Balance",
        wait_for_inclusion: bool = True,
        wait_for_finalization: bool = False,
    ) -> bool:
        nbnetwork_state = self.chain_state["NbNetworkModule"]

        bal = self.get_balance(wallet.coldkeypub.ss58_address)
        curr_stake = self.get_stake_for_coldkey_and_hotkey(
            hotkey_ss58=hotkey_ss58, coldkey_ss58=wallet.coldkeypub.ss58_address
        )
        if curr_stake is None:
            curr_stake = Balance(0)
        existential_deposit = self.get_existential_deposit()

        if bal < amount + existential_deposit:
            raise Exception("Insufficient funds")

        stake_state = nbnetwork_state["Stake"]

        # Stake the funds
        if not hotkey_ss58 in stake_state:
            stake_state[hotkey_ss58] = {}
        if not wallet.coldkeypub.ss58_address in stake_state[hotkey_ss58]:
            stake_state[hotkey_ss58][wallet.coldkeypub.ss58_address] = {}

        stake_state[hotkey_ss58][wallet.coldkeypub.ss58_address][
            self.block_number
        ] = amount.vim

        # Add to total_stake storage
        nbnetwork_state["TotalStake"][self.block_number] = (
            self._get_most_recent_storage(nbnetwork_state["TotalStake"])
            + amount.vim
        )

        total_hotkey_stake_state = nbnetwork_state["TotalHotkeyStake"]
        if not hotkey_ss58 in total_hotkey_stake_state:
            total_hotkey_stake_state[hotkey_ss58] = {}

        total_coldkey_stake_state = nbnetwork_state["TotalColdkeyStake"]
        if not wallet.coldkeypub.ss58_address in total_coldkey_stake_state:
            total_coldkey_stake_state[wallet.coldkeypub.ss58_address] = {}

        curr_total_hotkey_stake = self.query_nbnetwork(
            name="TotalHotkeyStake",
            params=[hotkey_ss58],
            block=min(self.block_number - 1, 0),
        )
        curr_total_coldkey_stake = self.query_nbnetwork(
            name="TotalColdkeyStake",
            params=[wallet.coldkeypub.ss58_address],
            block=min(self.block_number - 1, 0),
        )

        total_hotkey_stake_state[hotkey_ss58][self.block_number] = (
            curr_total_hotkey_stake.value + amount.vim
        )
        total_coldkey_stake_state[wallet.coldkeypub.ss58_address][
            self.block_number
        ] = (curr_total_coldkey_stake.value + amount.vim)

        # Remove from free balance
        self.chain_state["System"]["Account"][wallet.coldkeypub.ss58_address][
            "data"
        ]["free"][self.block_number] = (bal - amount).vim

        return True

    def _do_unstake(
        self,
        wallet: "wallet",
        hotkey_ss58: str,
        amount: "Balance",
        wait_for_inclusion: bool = True,
        wait_for_finalization: bool = False,
    ) -> bool:
        nbnetwork_state = self.chain_state["NbNetworkModule"]

        bal = self.get_balance(wallet.coldkeypub.ss58_address)
        curr_stake = self.get_stake_for_coldkey_and_hotkey(
            hotkey_ss58=hotkey_ss58, coldkey_ss58=wallet.coldkeypub.ss58_address
        )
        if curr_stake is None:
            curr_stake = Balance(0)

        if curr_stake < amount:
            raise Exception("Insufficient funds")

        stake_state = nbnetwork_state["Stake"]

        if curr_stake.vim == 0:
            return True

        # Unstake the funds
        # We know that the hotkey has stake, so we can just remove it
        stake_state[hotkey_ss58][wallet.coldkeypub.ss58_address][
            self.block_number
        ] = (curr_stake - amount).vim
        # Add to the free balance
        if (
            wallet.coldkeypub.ss58_address
            not in self.chain_state["System"]["Account"]
        ):
            self.chain_state["System"]["Account"][
                wallet.coldkeypub.ss58_address
            ] = {"data": {"free": {}}}

        # Remove from total stake storage
        nbnetwork_state["TotalStake"][self.block_number] = (
            self._get_most_recent_storage(nbnetwork_state["TotalStake"])
            - amount.vim
        )

        total_hotkey_stake_state = nbnetwork_state["TotalHotkeyStake"]
        if not hotkey_ss58 in total_hotkey_stake_state:
            total_hotkey_stake_state[hotkey_ss58] = {}
            total_hotkey_stake_state[hotkey_ss58][
                self.block_number
            ] = 0  # Shouldn't happen

        total_coldkey_stake_state = nbnetwork_state["TotalColdkeyStake"]
        if not wallet.coldkeypub.ss58_address in total_coldkey_stake_state:
            total_coldkey_stake_state[wallet.coldkeypub.ss58_address] = {}
            total_coldkey_stake_state[wallet.coldkeypub.ss58_address][
                self.block_number
            ] = amount.vim  # Shouldn't happen

        total_hotkey_stake_state[hotkey_ss58][self.block_number] = (
            self._get_most_recent_storage(
                nbnetwork_state["TotalHotkeyStake"][hotkey_ss58]
            )
            - amount.vim
        )
        total_coldkey_stake_state[wallet.coldkeypub.ss58_address][
            self.block_number
        ] = (
            self._get_most_recent_storage(
                nbnetwork_state["TotalColdkeyStake"][
                    wallet.coldkeypub.ss58_address
                ]
            )
            - amount.vim
        )

        self.chain_state["System"]["Account"][wallet.coldkeypub.ss58_address][
            "data"
        ]["free"][self.block_number] = (bal + amount).vim

        return True

    def get_delegate_by_hotkey(
        self, hotkey_ss58: str, block: Optional[int] = None
    ) -> Optional["DelegateInfo"]:
        nbnetwork_state = self.chain_state["NbNetworkModule"]

        if hotkey_ss58 not in nbnetwork_state["Delegates"]:
            return None

        newest_state = self._get_most_recent_storage(
            nbnetwork_state["Delegates"][hotkey_ss58], block
        )
        if newest_state is None:
            return None

        nom_result = []
        nominators = nbnetwork_state["Stake"][hotkey_ss58]
        for nominator in nominators:
            nom_amount = self.get_stake_for_coldkey_and_hotkey(
                hotkey_ss58=hotkey_ss58, coldkey_ss58=nominator, block=block
            )
            if nom_amount is not None and nom_amount.vim > 0:
                nom_result.append((nominator, nom_amount))

        registered_cosmos = []
        for cosmos in self.get_all_cosmos_netuids(block=block):
            uid = self.get_uid_for_hotkey_on_cosmos(
                hotkey_ss58=hotkey_ss58, netuid=cosmos, block=block
            )

            if uid is not None:
                registered_cosmos.append((cosmos, uid))

        info = DelegateInfo(
            hotkey_ss58=hotkey_ss58,
            total_stake=self.get_total_stake_for_hotkey(
                ss58_address=hotkey_ss58
            )
            or Balance(0),
            nominators=nom_result,
            owner_ss58=self.get_hotkey_owner(
                hotkey_ss58=hotkey_ss58, block=block
            ),
            take=0.18,
            validator_permits=[
                cosmos
                for cosmos, uid in registered_cosmos
                if self.particle_has_validator_permit(
                    uid=uid, netuid=cosmos, block=block
                )
            ],
            registrations=[cosmos for cosmos, _ in registered_cosmos],
            return_per_1000=Balance.from_nim(
                1234567
            ),  # Doesn't matter for mock?
            total_daily_return=Balance.from_nim(
                1234567
            ),  # Doesn't matter for mock?
        )

        return info

    def get_delegates(
        self, block: Optional[int] = None
    ) -> List["DelegateInfo"]:
        nbnetwork_state = self.chain_state["NbNetworkModule"]
        delegates_info = []
        for hotkey in nbnetwork_state["Delegates"]:
            info = self.get_delegate_by_hotkey(hotkey_ss58=hotkey, block=block)
            if info is not None:
                delegates_info.append(info)

        return delegates_info

    def get_delegated(
        self, coldkey_ss58: str, block: Optional[int] = None
    ) -> List[Tuple["DelegateInfo", "Balance"]]:
        """Returns the list of delegates that a given coldkey is staked to."""
        delegates = self.get_delegates(block=block)

        result = []
        for delegate in delegates:
            if coldkey_ss58 in delegate.nominators:
                result.append((delegate, delegate.nominators[coldkey_ss58]))

        return result

    def get_all_cosmos_info(
        self, block: Optional[int] = None
    ) -> List[CosmosInfo]:
        nbnetwork_state = self.chain_state["NbNetworkModule"]
        result = []
        for cosmos in nbnetwork_state["NetworksAdded"]:
            info = self.get_cosmos_info(netuid=cosmos, block=block)
            if info is not None:
                result.append(info)

        return result

    def get_cosmos_info(
        self, netuid: int, block: Optional[int] = None
    ) -> Optional[CosmosInfo]:
        if not self.cosmos_exists(netuid=netuid, block=block):
            return None

        def query_cosmos_info(name: str) -> Optional[object]:
            return self.query_nbnetwork(
                name=name, block=block, params=[netuid]
            ).value

        info = CosmosInfo(
            netuid=netuid,
            rho=query_cosmos_info(name="Rho"),
            kappa=query_cosmos_info(name="Kappa"),
            difficulty=query_cosmos_info(name="Difficulty"),
            immunity_period=query_cosmos_info(name="ImmunityPeriod"),
            max_allowed_validators=query_cosmos_info(
                name="MaxAllowedValidators"
            ),
            min_allowed_weights=query_cosmos_info(name="MinAllowedWeights"),
            max_weight_limit=query_cosmos_info(name="MaxWeightLimit"),
            scaling_law_power=query_cosmos_info(name="ScalingLawPower"),
            cosmos_n=query_cosmos_info(name="CosmosN"),
            max_n=query_cosmos_info(name="MaxAllowedUids"),
            blocks_since_epoch=query_cosmos_info(name="BlocksSinceLastStep"),
            tempo=query_cosmos_info(name="Tempo"),
            modality=query_cosmos_info(name="NetworkModality"),
            connection_requirements={
                str(netuid_.value): percentile.value
                for netuid_, percentile in self.query_map_nbnetwork(
                    name="NetworkConnect", block=block, params=[netuid]
                ).records
            },
            emission_value=query_cosmos_info(name="EmissionValues"),
            burn=query_cosmos_info(name="Burn"),
            owner_ss58=query_cosmos_info(name="CosmosOwner"),
        )

        return info

    def _do_serve_proton(
        self,
        wallet: "wallet",
        call_params: "ProtonServeCallParams",
        wait_for_inclusion: bool = False,
        wait_for_finalization: bool = True,
    ) -> Tuple[bool, Optional[str]]:
        return True, None

    def _do_set_weights(
        self,
        wallet: "wallet",
        netuid: int,
        uids: int,
        vals: List[int],
        version_key: int,
        wait_for_inclusion: bool = False,
        wait_for_finalization: bool = True,
    ) -> Tuple[bool, Optional[str]]:
        return True, None

    def _do_serve_fermion(
        self,
        wallet: "wallet",
        call_params: "FermionServeCallParams",
        wait_for_inclusion: bool = False,
        wait_for_finalization: bool = True,
    ) -> Tuple[bool, Optional[str]]:
        return True, None
